/*
 * 3d.c in Virtual Combat
 * Copyright (c) 1993 by David Jameson
 *
 * Permission to use, copy, modify and distribute this software
 * for any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission appear in the supporting
 * documentation.
 */


/* 3d routines */

#define DIFFERENT 0  /* Used for line of sight calculation, indicates that
			a polygon is not in LOS */
#define COLLIDEZ 1   /* Collision in the z direction i.e. player has walked
			into an object */
#define COLLIDEY 2   /* Collision in the y direction i.e. player has hit
			the ground */

#include "defs.h"
#include "3d.h"

#include <math.h>
#include <limits.h>
#include <values.h>

/* Module variables */

static disp *displ;    /* Display information */
static long scale;
static long original_x, original_z;
static int y_collisions;    /* Number of collisions in the y direction */
static int z_collisions;    /* Number of collisions in the z direction */
static long centrex, centrey;
static point teleportdata;
static char *collidedata;    /* Message sent to server when a collision
				occurs */
static char *portaldata;     /* Message sent to server when a collision
				occurs with portal object */


/* Function prototypes */

static int compare(const void *entry1, const void *entry2);



/* Routine to setup a unit matrix */
void unit_matrix(matrix *mat)
{
    mat->a1 = 1;
    mat->b2 = 1;
    mat->c3 = 1;
    mat->a2 = 0;
    mat->a3 = 0;
    mat->b1 = 0;
    mat->b3 = 0;
    mat->c1 = 0;
    mat->c2 = 0;
} /* unit_matrix */


/* Spherical polar to cartesian conversion */
void polar_to_xyz(
	point *result,
	float dist,	/* Length of vector */
	float elev,	/* Elevation above horizontal */
	float yrot)	/* rotation about y-axis */
{
    float horiz;

    result->y = dist * sin(elev);
    horiz = dist * cos(elev);
    result->x = horiz * sin(yrot);
    result->z = horiz * cos(yrot);
}  /* polar_to_xyz() */


/* Make an object a child of another object */
void add_child(object *child, object *obj)
{
    ++obj->children;
    if (! obj->childobj) {
	if (! (obj->childobj = (object **)malloc(obj->children *
	  sizeof(object *)))) {
	     perror("Error allocating memory for object pointer list");
	     exit(1);
	}
    }

    if (! (obj->childobj = (object **)realloc((char *)obj->childobj,
      obj->children * sizeof(object *)))) {
	perror("Error reallocating memory for object pointer list");
	exit(1);
    }
    child_object(obj, obj->children - 1) = child;
    child->parentobj = obj;
}  /* add_child() */


/* Remove a child object from its parent object */
void delete_child(object *obj, int number)
{
    --obj->children;
    if (! obj->children) free(obj->childobj);
    else {
	/* Remove entry from array */
	child_object(obj, number) = child_object(obj, obj->children);
	if (! (obj->childobj = (object **)realloc((char *)obj->childobj,
	  obj->children * sizeof(object *)))) {
	    fprintf(stderr,
		"Error reallocating memory for object pointer list\n");
	    exit(1);
	}
    }
}  /* delete_child() */


/* Routine to multiply two 3x3 matrices together */
void matmult(matrix *mat1, matrix *mat2, matrix *result)
{
    matrix tempmat;

    tempmat.a1 = (mat1->a1*mat2->a1)+(mat1->a2*mat2->b1)+(mat1->a3*mat2->c1);
    tempmat.a2 = (mat1->a1*mat2->a2)+(mat1->a2*mat2->b2)+(mat1->a3*mat2->c2);
    tempmat.a3 = (mat1->a1*mat2->a3)+(mat1->a2*mat2->b3)+(mat1->a3*mat2->c3);

    tempmat.b1 = (mat1->b1*mat2->a1)+(mat1->b2*mat2->b1)+(mat1->b3*mat2->c1);
    tempmat.b2 = (mat1->b1*mat2->a2)+(mat1->b2*mat2->b2)+(mat1->b3*mat2->c2);
    tempmat.b3 = (mat1->b1*mat2->a3)+(mat1->b2*mat2->b3)+(mat1->b3*mat2->c3);

    tempmat.c1 = (mat1->c1*mat2->a1)+(mat1->c2*mat2->b1)+(mat1->c3*mat2->c1);
    tempmat.c2 = (mat1->c1*mat2->a2)+(mat1->c2*mat2->b2)+(mat1->c3*mat2->c2);
    tempmat.c3 = (mat1->c1*mat2->a3)+(mat1->c2*mat2->b3)+(mat1->c3*mat2->c3);

    *result = tempmat;
}  /* matmult */


/* Routine to rotate a 3-d point using a 3x3 matrix */
void rotpoint(matrix *mat, point *pt)
{
    /* This routine will be called a lot of times, so make variables
       static so that memory doesn't have to be allocated each time */

    static point temppt;

    temppt.x = (mat->a1 * pt->x) + (mat->a2 * pt->y) + (mat->a3 * pt->z);
    temppt.y = (mat->b1 * pt->x) + (mat->b2 * pt->y) + (mat->b3 * pt->z);
    temppt.z = (mat->c1 * pt->x) + (mat->c2 * pt->y) + (mat->c3 * pt->z);

    *pt = temppt;
} /* rotpoint() */


/* Find composite orientation matrix for a hierarchical object by
   traversing the tree structure backwards */
void composite_matrix(
	object *obj,  /* Object that we are working on */
	sortentry *entry)  /* Pointer to entry in sort-table */
{
    entry->displ = obj->centre;
    entry->mat = obj->rotmatrix;
    while (obj->parentobj) {
	obj = obj->parentobj;
	matmult(&obj->rotmatrix, &entry->mat, &entry->mat);
	rotpoint(&obj->rotmatrix, &entry->displ);
	entry->displ.x += obj->centre.x;
	entry->displ.y += obj->centre.y;
	entry->displ.z += obj->centre.z;
    }
    /* Rotate relative to viewer */
    matmult(&viewer.inversematrix, &entry->mat, &entry->mat);
    /* Now get displacement relative to viewer */
    entry->relative.x = entry->displ.x - viewer.position.x;
    entry->relative.y = entry->displ.y - viewer.position.y;
    entry->relative.z = entry->displ.z - viewer.position.z;
    rotpoint(&viewer.inversematrix, &entry->relative);

}  /* composite_matrix() */


/* Rotate object about x-axis */
void rotate_x(
	matrix *mat,  /* Orientation matrix for object */
	float angle,  /* Angle to rotate (radians) */
	int mode)     /* Either WORLDMODE or LOCALMODE */
{
    float cosang, sinang;
    static matrix rotmat;

    cosang = (float)cos((double)angle);
    sinang = (float)sin((double)angle);
    rotmat.a1 = 1;
    rotmat.a2 = 0;
    rotmat.a3 = 0;
    rotmat.b1 = 0;
    rotmat.b2 = cosang;
    rotmat.b3 = -sinang;
    rotmat.c1 = 0;
    rotmat.c2 = sinang;
    rotmat.c3 = cosang;

    /* If WORLDMODE, we rotate about world axes.  This means we rotate
     * last, so we must put the rotation matrix FIRST.
     * Otherwise, we put rotation matrix after the object's matrix.
     */

    if (mode == WORLDMODE) matmult(&rotmat, mat, mat);
    else matmult(mat, &rotmat, mat);

    /* If not WORLDMODE, let's assume mode==LOCALMODE - we don't have
     * time to check.
     */
}  /* rotate_x */


/* Rotate object about y-axis */
void rotate_y(
	matrix *mat,  /* Orientation matrix for object */
	float angle,  /* Angle to rotate (radians) */
	int mode)     /* Either WORLDMODE or LOCALMODE */
{
    float cosang, sinang;
    static matrix rotmat;

    cosang = (float)cos((double)angle);
    sinang = (float)sin((double)angle);
    rotmat.b2 = 1;
    rotmat.b3 = 0;
    rotmat.b1 = 0;
    rotmat.c2 = 0;
    rotmat.c3 = cosang;
    rotmat.c1 = -sinang;
    rotmat.a2 = 0;
    rotmat.a3 = sinang;
    rotmat.a1 = cosang;

    /* If WORLDMODE, we rotate about world axes.  This means we rotate
     * last, so we must put the rotation matrix FIRST.
     * Otherwise, we put rotation matrix after the object's matrix.
     */

    if (mode == WORLDMODE) matmult(&rotmat, mat, mat);
    else matmult(mat, &rotmat, mat);

    /* If not WORLDMODE, let's assume mode==LOCALMODE - we don't have
     * time to check.
     */
}  /* rotate_y */


/* Rotate object about z-axis */
void rotate_z(
	matrix *mat,  /* Orientation matrix for object */
	float angle,  /* Angle to rotate (radians) */
	int mode)     /* Either WORLDMODE or LOCALMODE */
{
    float cosang, sinang;
    static matrix rotmat;

    cosang = (float)cos((double)angle);
    sinang = (float)sin((double)angle);
    rotmat.c3 = 1;
    rotmat.c1 = 0;
    rotmat.c2 = 0;
    rotmat.a3 = 0;
    rotmat.a1 = cosang;
    rotmat.a2 = -sinang;
    rotmat.b3 = 0;
    rotmat.b1 = sinang;
    rotmat.b2 = cosang;

    /* If WORLDMODE, we rotate about world axes.  This means we rotate
     * last, so we must put the rotation matrix FIRST.
     * Otherwise, we put rotation matrix after the object's matrix.
     */

    if (mode == WORLDMODE) matmult(&rotmat, mat, mat);
    else matmult(mat, &rotmat, mat);

    /* If not WORLDMODE, let's assume mode==LOCALMODE - we don't have
     * time to check.
     */
}  /* rotate_z */


/* Routine to rotate points relative to viewer */
int rotate_points(
	sortentry *entry,     /* Object to be rotated (entry in sort-table) */
	point *pointdata)     /* Storage area for rotated points */
{
    object *obj;
    point *objpoints;
    int object_visible;
    point pt;
    int i;

    obj = entry->objptr;
    objpoints = obj->points;
    object_visible = FALSE;  /* Object invisible (until shown otherwise) */

    for (i=0; i < obj->no_points; i++) {
	pt.x = objpoints[i].x;
	pt.y = objpoints[i].y;
	pt.z = objpoints[i].z;
	rotpoint(&entry->mat, &pt);
	pointdata[i].x = pt.x + entry->relative.x;
	pointdata[i].y = pt.y + entry->relative.y;

	if ((pointdata[i].z = pt.z + entry->relative.z) > CLIPDIST)
	    object_visible = TRUE;
    }
    return object_visible;
}  /* rotate_points */


/* Routine to do perspective on rotated points */
void do_perspective(
	point *pointdata,
	screenpoint *screendata,
	int no_points)
{
    register int i;

    for (i=0; i < no_points; i++) {
	if (pointdata[i].z > CLIPDIST) {
	    screendata[i].x = (int)((pointdata[i].x * scale) /
				     pointdata[i].z);
	    screendata[i].y = (int)((pointdata[i].y * scale) /
				     pointdata[i].z);
	}
	else {
	    screendata[i].x = 0;
	    screendata[i].y = 0;
	}
    }
}  /* do_perspective */


/* Draw each surface of object on screen */
void draw_object(
	point *rotpoints,  /* Array of rotated points (before perspective)
			      (actual x, y and z coords relative to viewer) */
	screenpoint *screenpoints,  /* Array of points used for actually
			  drawing polygons on screen (x and y screen coords) */
	object *obj,  /* Object to be drawn */
	long *LOS_z,  /* Z coordinate of polygon in line of sight which is
			 nearest to the player */
	object **LOS_obj)  /* Object which is in LOS */
{
    int i, j;
    int surfptr;  /* Pointer into surface table */
    point *p1, *p2, *p3;
    point d1, d2;
    float nx, ny, nz, dotproduct;
    unsigned short *surf;
    Drawable dr;
    XPoint xpt[MAXSIZEPOLY];
    int vertices;  /* Number of vertices in polygon we are dealing with */
    int xptptr;  /* Pointer into xpt array */
    float temp;
    point *coord, *nextcoord;  /* For collision detection */
    int clxsurf;  /* Surface where last collision detected (if any) */
    float light;  /* Intensity of colour reflecting off object */
    int shade;	  /* Shade of colour used (brightness) */
    int zsign;
    int ysign;
    float crossz;
    float crossy;
    float isectz;  /* z-coord of intersection with line x=0,y=0 */
    float isecty;
    float elev;
    int do_collision;
    float tempx, tempy;


    clxsurf = 0;
    surfptr = 0;
    surf = obj->surface;
    for (i=0; i < (obj->no_surfaces); i++) {
	float len;

	/* First see if we need to draw this surface */

	p1 = &rotpoints[surf[surfptr + 1]];
	p2 = &rotpoints[surf[surfptr + 2]];
	p3 = &rotpoints[surf[surfptr + 3]];

	d1.x = p2->x - p1->x;
	d2.x = p3->x - p2->x;
	d1.y = p2->y - p1->y;
	d2.y = p3->y - p2->y;
	d1.z = p2->z - p1->z;
	d2.z = p3->z - p2->z;

	/* Calculate normal to surface (d2 X d1) */

	nx = (float)d2.y * (float)d1.z - (float)d2.z * (float)d1.y;
	ny = (float)d2.z * (float)d1.x - (float)d2.x * (float)d1.z;
	nz = (float)d2.x * (float)d1.y - (float)d2.y * (float)d1.x;

	/* Calculate square of tangent of elevation of surface normal.
	   If elev > 1.0 then elevation of polygon is less than 45 degrees
	   (used for collision detection) */

	if (nx == 0.0 && nz == 0.0) elev = 10.0;
	else elev = ny * ny / (nx * nx + nz * nz);

	/* Find dotproduct of surface normal with vector between viewer
	   and surface */

	dotproduct = p1->x * nx + p1->y * ny + p1->z * nz;

	/* Calculate intensity of colour by calculating the cosine of the
	   angle at which the surface is facing the player and then
	   multiplying by the relative intensity of the colour.  Surfaces
	   facing the player directly will be brighter i.e. the player is
	   the light source. */

	len = sqrt((double)((float)p1->x * p1->x + (float)p1->y * p1->y
	      + (float)p1->z * p1->z) * (nx * nx + ny * ny + nz * nz));
	if (len > 0.001) light = fabs((double)dotproduct) / len;
	else light = 0.0;
	light *= obj->light / 2;
	light += obj->light / 2;

	vertices = (int)surf[surfptr];
	surfptr++;
	if ((obj->flags & ALWAYS_VISIBLE) || (dotproduct < 0.0 )) {
	    /* surface visible */
	    xptptr = 0;  /* Pointer into our xpt[] array, used to store the
			    final x and y values of the vertices of the
			    polygon being drawn. */

	    for (j = 0; j < vertices; j++) {
		/* Check to see if polygon is in line of sight with
		   gunsights by seeing if the polygon encloses the point
		   (x=0, y=0) */
		coord = &rotpoints[surf[surfptr + j]];
		if (j != (vertices - 1))
		    nextcoord = &rotpoints[surf[surfptr + j + 1]];
		else nextcoord = &rotpoints[surf[surfptr]];
		/* Find z coordinate of cross-product between
		   one side of polygon and line joining vertex to
		   the origin.
		   If the sign of the z coordinte changes then the
		   polygon doesn't enclose (x=0,y=0). */
		crossz = (nextcoord->x - coord->x) * (float)coord->y -
			 (nextcoord->y - coord->y) * (float)coord->x;
		/* Find y coordinate of cross-product, for collision
		   detection */
		crossy = (nextcoord->z - coord->z) * (float)coord->x -
			 (nextcoord->x - coord->x) * (float)coord->z;
		if (j == 0) {
		    zsign = sign(crossz);
		    ysign = sign(crossy);
		}
		else {
		    if (sign(crossz) != zsign) zsign = DIFFERENT;
		    if (sign(crossy) != ysign) ysign = DIFFERENT;
		}
		/*
		 * Now check this point with next point to see if
		 * a collision has occurred (unless player hasn't moved
		 * position since last time).
		 * No collision detection is done for player objects
		 * (it is handled by the client).
		 */
		if ((! (obj->flags & PLAYER)) &&
		    (min(coord->y, nextcoord->y) <= CLXHEIGHT) &&
		    (max(coord->y, nextcoord->y) >= 0) &&
		    (viewer.yvel || viewer.zvel) && (/*surfptr !=*/ !clxsurf) &&
		    ((y_collisions == 0) || (elev < 1.0))) {
		    /* Y coordinate has indicated a possible collision.
		       Check to see if any side of the polygon
		       intersects a cylindrical boundary around the
		       player */
		    do_collision = FALSE;

		    if (((float)coord->x * coord->x + (float)coord->z *
		      coord->z < CLXRADIUS * CLXRADIUS) &&
		      (coord->y > 0) && (coord->y <= CLXHEIGHT))
			do_collision = TRUE;
		    else if (nextcoord->x == coord->x && nextcoord->z ==
		      coord->z && ((float)coord->x * coord->x +
		      (float)coord->z * coord->z < CLXRADIUS * CLXRADIUS))
			do_collision = TRUE;
		    else {
			float dx, dz, t, newx, newz;

			dx = nextcoord->x - coord->x;
			dz = nextcoord->z - coord->z;
			/* Calculate the point where the projection of the
			   line on the x-z plane comes closest to (0,0). */
			t = -(dx * coord->x + dz * coord->z) /
			     (dx * dx + dz * dz);
			/* t is how far along the line the closest
			   point is */
			if ((t > 0.0) && (t < 1.0)) {
			    newx = coord->x + t * dx;
			    newz = coord->z + t * dz;
			    if ((newx * newx + newz * newz) <
			      CLXRADIUS * CLXRADIUS) do_collision = TRUE;
			}
		    }
		    if (do_collision) {
			if (elev >= 1.0) {
			    float isecty;
			 /*   float factor;
			    point displ;*/

			    isecty = (p1->x * nx + p1->z * nz) /
				     ny + p1->y;
			    if ((isecty >= 0) && (isecty < CLXHEIGHT)) {
				y_collisions++;
				clxsurf = surfptr;
				viewer.position.y -=
				    CLXHEIGHT - isecty;
			     }


			    /* Viewer has hit polygon, so calculate
			       rebound displacement */

		  /*	      factor = -(viewer.yvel * ny +
				       viewer.zvel * nz) /
				      (nx * nx + ny * ny + nz * nz);
			    displ.x = nx * factor;
			    displ.y = ny * factor;
			    displ.z = nz * factor;
			    rotpoint(&viewer.rotmatrix, &displ);
			    viewer.position.x += displ.x;
			    viewer.position.y += displ.y;
			    viewer.position.z += displ.z;*/
			}
			/*
			 * Only check for collisions in the z direction
			 * if player is walking towards polygon.  This
			 * stops players getting stuck inside objects.
			 */
			else if ((sign(nz) != sign(viewer.zvel)) ||
			         (obj->flags & ALWAYS_VISIBLE)) {
			    if (z_collisions == 1) {
				/*
				 * Restore viewer's x and z coords to
				 * what they were before the player
				 * started moving.
				 */
				point displ;

				z_collisions++;
				viewer.position.x = original_x;
				viewer.position.z = original_z;
				displ.x = displ.y = 0;
				displ.z = -viewer.zvel * 1.1;
				rotpoint(&viewer.rotmatrix, &displ);
				viewer.position.x += displ.x;
				viewer.position.z += displ.z;
			    }
			    else if (z_collisions == 0) {
				float factor;
				point displ;

				z_collisions++;
				clxsurf = surfptr;
				/* Viewer has hit polygon, so calculate
				   rebound displacement */

				factor = -(viewer.yvel * ny +
					   viewer.zvel * nz) /
					  (nx * nx + ny * ny + nz * nz);
				displ.x = nx * factor;
				displ.y = ny * factor;
				displ.z = nz * factor;

				displ.z -= viewer.zvel * 1.1;
				rotpoint(&viewer.rotmatrix, &displ);
				viewer.position.x += displ.x;
				viewer.position.y += displ.y;
				viewer.position.z += displ.z;
			    }
			}
		    }
		}

		if (rotpoints[surf[surfptr + j]].z <= CLIPDIST) {

		    /* Check with previous point to see if we need to do a
		       z-clip i.e. check to see if the line joining the two
		       points goes through the plane z=CLIPDIST */

		    float x1, x2, y1, y2, z1, z2;

		    if (j > 0) if (rotpoints[surf[surfptr + j - 1]].z >
				   CLIPDIST) {
			x1 = (float)rotpoints[surf[surfptr + j]].x;
			y1 = (float)rotpoints[surf[surfptr + j]].y;
			z1 = (float)rotpoints[surf[surfptr + j]].z;
			x2 = (float)rotpoints[surf[surfptr + j - 1]].x;
			y2 = (float)rotpoints[surf[surfptr + j - 1]].y;
			z2 = (float)rotpoints[surf[surfptr + j - 1]].z;
			temp = (CLIPDIST - z1) / (z2 - z1);
			tempx = (((x2 - x1) * temp + x1) *
					      scale / CLIPDIST) + centrex;
			tempy = (((y2 - y1) * temp + y1) *
					      scale / CLIPDIST) + centrey;
			if (tempx > SHRT_MAX) tempx = SHRT_MAX;
			if (tempx < SHRT_MIN) tempx = SHRT_MIN;
			if (tempy > SHRT_MAX) tempy = SHRT_MAX;
			if (tempy < SHRT_MIN) tempy = SHRT_MIN;
			xpt[xptptr].x = tempx;
			xpt[xptptr].y = tempy;
			xptptr++;
		    }
		    if (j == 0) if (rotpoints[surf[surfptr + vertices - 1]].z
				    > CLIPDIST) {
			x1 = (float)rotpoints[surf[surfptr + j]].x;
			y1 = (float)rotpoints[surf[surfptr + j]].y;
			z1 = (float)rotpoints[surf[surfptr + j]].z;
			x2 = (float)rotpoints[surf[surfptr + vertices - 1]].x;
			y2 = (float)rotpoints[surf[surfptr + vertices - 1]].y;
			z2 = (float)rotpoints[surf[surfptr + vertices - 1]].z;
			temp = (CLIPDIST - z1) / (z2 - z1);
			tempx = (((x2 - x1) * temp + x1) *
					      scale / CLIPDIST) + centrex;
			tempy = (((y2 - y1) * temp + y1) *
					      scale / CLIPDIST) + centrey;
			if (tempx > SHRT_MAX) tempx = SHRT_MAX;
			if (tempx < SHRT_MIN) tempx = SHRT_MIN;
			if (tempy > SHRT_MAX) tempy = SHRT_MAX;
			if (tempy < SHRT_MIN) tempy = SHRT_MIN;
			xpt[xptptr].x = tempx;
			xpt[xptptr].y = tempy;
			xptptr++;
		    }
		    /* Now check with next point */
		    if (j != (vertices - 1))
			if (rotpoints[surf[surfptr + j + 1]].z > CLIPDIST) {
			x1 = (float)rotpoints[surf[surfptr + j]].x;
			y1 = (float)rotpoints[surf[surfptr + j]].y;
			z1 = (float)rotpoints[surf[surfptr + j]].z;
			x2 = (float)rotpoints[surf[surfptr + j + 1]].x;
			y2 = (float)rotpoints[surf[surfptr + j + 1]].y;
			z2 = (float)rotpoints[surf[surfptr + j + 1]].z;
			temp = (CLIPDIST - z1) / (z2 - z1);
			tempx = (((x2 - x1)*temp + x1) *
					      scale / CLIPDIST) + centrex;
			tempy = (((y2 - y1)*temp + y1) *
					      scale / CLIPDIST) + centrey;
			if (tempx > SHRT_MAX) tempx = SHRT_MAX;
			if (tempx < SHRT_MIN) tempx = SHRT_MIN;
			if (tempy > SHRT_MAX) tempy = SHRT_MAX;
			if (tempy < SHRT_MIN) tempy = SHRT_MIN;
			xpt[xptptr].x = tempx;
			xpt[xptptr].y = tempy;
			xptptr++;
		    }
		    if (j == (vertices - 1))
			if (rotpoints[surf[surfptr]].z > CLIPDIST) {
			x1 = (float)rotpoints[surf[surfptr + j]].x;
			y1 = (float)rotpoints[surf[surfptr + j]].y;
			z1 = (float)rotpoints[surf[surfptr + j]].z;
			x2 = (float)rotpoints[surf[surfptr]].x;
			y2 = (float)rotpoints[surf[surfptr]].y;
			z2 = (float)rotpoints[surf[surfptr]].z;
			temp = (CLIPDIST - z1) / (z2 - z1);
			/* Still need to do perspective on clipped point
			   (z coord = CLIPDIST) */
			tempx = (((x2 - x1)*temp + x1) *
					      scale / CLIPDIST) + centrex;
			tempy = (((y2 - y1)*temp + y1) *
					      scale / CLIPDIST) + centrey;
			if (tempx > SHRT_MAX) tempx = SHRT_MAX;
			if (tempx < SHRT_MIN) tempx = SHRT_MIN;
			if (tempy > SHRT_MAX) tempy = SHRT_MAX;
			if (tempy < SHRT_MIN) tempy = SHRT_MIN;
			xpt[xptptr].x = tempx;
			xpt[xptptr].y = tempy;
			xptptr++;
		    }
		}  /* End of z-clipping */
		else {
		    tempx = screenpoints[surf[surfptr + j]].x +
				    centrex;
		    tempy = screenpoints[surf[surfptr + j]].y +
				    centrey;
		    if (tempx > SHRT_MAX) tempx = SHRT_MAX;
		    if (tempx < SHRT_MIN) tempx = SHRT_MIN;
		    if (tempy > SHRT_MAX) tempy = SHRT_MAX;
		    if (tempy < SHRT_MIN) tempy = SHRT_MIN;
		    xpt[xptptr].x = tempx;
		    xpt[xptptr].y = tempy;
		    xptptr++;
		}
	    }
	    /* Do a second collision detection, for the whole polygon
	       (instead of just its sides) :
	       Check to see if polygon intersects the line
	       x=0,z=0 within the range 0 < y < CLXHEIGHT */
	    if ((ysign != DIFFERENT) && (/*surfptr !=*/ ! clxsurf)) {
		/* Calculate the y-coordinate of the point where the polygon
		   intersects with the line x=0,z=0 */

		isecty = (p1->x * nx + p1->z * nz) / ny + p1->y;
		if ((isecty >= 0) && (isecty < CLXHEIGHT)) {
		    y_collisions++;
		    clxsurf = surfptr;
		    viewer.position.y -= (CLXHEIGHT - isecty);
		}
	    }

	    if (zsign != DIFFERENT) {
		/* Calculate the z-coordinate of the point where the polygon
		   intersects with gunsight (the line x=0,y=0) */

		isectz = (p1->x * nx + p1->y * ny) / nz + p1->z;

		/* If object is in line of sight, check to see whether there
		   are any nearer polygons also in LOS. */
		if ((isectz > 0) && ((isectz < *LOS_z) || (*LOS_z < 0))) {
		    *LOS_z = isectz;
		    *LOS_obj = obj;
		}
	    }
	    if (xptptr > 2) {
		dr = (Drawable)displ->pix;
		/* Calculate closest shade, adding 0.5 to round to nearest
		   integer */
		shade = (int)(light * (SHADES - 1) + 0.5);
		if ((shade > SHADES - 1) || (shade < 0)) {
		    fprintf(stderr, "Error: illegal colour.\n");
		    exit(1);
		}
		XSetForeground(displ->dpy, displ->gc,
			       obj->col->shade[shade].pixel);
		XFillPolygon(displ->dpy, dr, displ->gc, xpt, xptptr,
			     Convex, CoordModeOrigin);
	    }
	}
	surfptr += (int)surf[surfptr - 1];
    }
    /*
     * If a collision has occurred then check to see if there is a
     * collision/portal string associated with the object.
     */
    if (clxsurf && obj->collide) collidedata = obj->collide;
    if (clxsurf && obj->portal) {
	portaldata = obj->portal;
        teleportdata = obj->teleport;
    }
}  /* draw_object */


/* Calculate the distance between two points */
float distance(point *point1, point *point2)
{
    float dist, dx, dy, dz;

    dx = (float)(point1->x - point2->x);
    dy = (float)(point1->y - point2->y);
    dz = (float)(point1->z - point2->z);

    dist = (float)sqrt((double)(dx * dx + dy * dy + dz * dz));
    return dist;
}  /* distance() */


/* Add an object (or hierarchical list of objects) to sort-list */
int add_sort(
	object *obj,
	sortentry *table,  /* Current position in table */
	int maxsize) /* Max table entries (from current position
			in the table) */
{
    float dist;
    int entries;  /* Number of entries we've added to table */
    double dx, dy, dz;
    int i;

    entries = 0;

    /* If colour not set then inherit from parent */
    if (! obj->col) {
	obj->col = obj->parentobj->col;
	obj->light = obj->parentobj->light;
    }
    composite_matrix(obj, &table[entries]);
		       /* This could be more efficient */
    if (entries == maxsize) {
	fprintf(stderr, "Error: sort table overflow.\n");
	exit(1);
    }
    dx = (double)table[entries].relative.x;
    dy = (double)table[entries].relative.y;
    dz = (double)table[entries].relative.z;

    /* Calculate how far viewer is from object */
    dist = (float)sqrt(dx * dx + dy * dy  + dz * dz);
    /* If object is too far to be seen and a less detailed object has
       been defined, use the less detailed object */
    while ((dist > (float)(obj->maxdist)) && (obj->farobj))
	obj = obj->farobj;
    if (dist <= (float)(obj->maxdist)) {
	if (obj->flags & DRAW_LAST) table[entries].distance = 0;
	else if (obj->flags & DRAW_FIRST)
	    table[entries].distance = MAXFLOAT;
	else table[entries].distance = dist;
	table[entries].objptr = obj;
	++entries;
	for (i = 0; i < obj->children; i++)
	    entries += add_sort(child_object(obj, i), table + entries,
		    maxsize - entries);
    }
    return entries;
}  /* add_sort */


/* compare routine needed for qsort() */
static int compare(const void *entry1, const void *entry2)
{
    /* We need array sorted in descending order
       (i.e. furthest object first) */
    static sortentry *ptr1, *ptr2;

    ptr1 = (sortentry *)entry1;
    ptr2 = (sortentry *)entry2;
    if (ptr1->distance == ptr2->distance) return 0;
    if (ptr1->distance > ptr2->distance) return -1;
    return 1;
}  /* compare() */



/* Depth-sort all objects to be drawn (furthest away drawn first) */
void depth_sort(sortentry *table, int size)
{
    int tablepos, check;
    sortentry tempentry;
    long planecoord, objcoord, viewercoord;

    /* Not very efficient */
    qsort((void *)table, size, sizeof(sortentry), compare);

    /* Now check to see if any plane needs to be moved back in the table
       (i.e. should be drawn sooner) */
    for (check = 1; check < size; check++) {
	if (table[check].objptr->flags & CHECKPOLY) {
	    for (tablepos = 0; tablepos < check; tablepos++) {
		if (table[tablepos].objptr->flags & CHECKOBJ) {
		    switch (table[check].objptr->flags & CHECKPOLY) {
			case CHECKX :
			    planecoord = table[check].displ.x;
			    objcoord = table[tablepos].displ.x;
			    viewercoord = viewer.position.x;
			    break;
			case CHECKY :
			    planecoord = table[check].displ.y;
			    objcoord = table[tablepos].displ.y;
			    viewercoord = viewer.position.y;
			    break;
			case CHECKZ :
			    planecoord = table[check].displ.z;
			    objcoord = table[tablepos].displ.z;
			    viewercoord = viewer.position.z;
			    break;
			default :
			    fprintf(stderr,
			      "Illegal CHECKPOLY flag combination.\n");
			    exit(1);
		    }
		    /*
		     * Plane needs to be moved towards start of table if
		     * object and viewer are on the same side of the plane.
		     */
		    if (sign(objcoord - planecoord) ==
			sign(viewercoord - planecoord)) {
			/* Do another check to see if it's really necessary
			   to reposition the objects.  See if the smaller
			   object is near the plane object in terms of the
			   two plane coordinates. */
			int i;
			point *points, temp;
			long  maxx, maxy, maxz;

			maxx = maxy = maxz = 0L;
			points = table[check].objptr->points;
			for (i = 0; i < table[check].objptr->no_points; i++) {
			    temp = points[i];
			    rotpoint(&table[check].objptr->rotmatrix,
				     &temp);
			    if (labs(temp.x) > maxx)
				maxx = labs(temp.x);
			    if (labs(temp.y) > maxy)
				maxy = labs(temp.y);
			    if (labs(temp.z) > maxz)
				maxz = labs(temp.z);
			}
			/* If the object being checked is outside the area
			   covered by the plane object then continue
			   inner loop */
			if (table[check].objptr->flags & CHECKX) {
			    if (labs(table[tablepos].displ.y -
				     table[check].displ.y) > maxy) continue;
			    if (labs(table[tablepos].displ.z -
				     table[check].displ.z) > maxz) continue;
			}
			else if (table[check].objptr->flags & CHECKY) {
			    if (labs(table[tablepos].displ.x -
				     table[check].displ.x) > maxx) continue;
			    if (labs(table[tablepos].displ.z -
				     table[check].displ.z) > maxz) continue;
			}
			else {	/* CHECKZ must be set */
			    if (labs(table[tablepos].displ.x -
				     table[check].displ.x) > maxx) continue;
			    if (labs(table[tablepos].displ.y -
				     table[check].displ.y) > maxy) continue;
			}

			tempentry = table[check];
			bcopy((char *)&table[tablepos],
				(char *)&table[tablepos + 1],
				(check - tablepos) * sizeof(sortentry));
			table[tablepos] = tempentry;
			break;
		    }
		}
	    }
	}
    }

    /* Now check to see if any plane needs to be moved forward in the table
       (i.e. should be drawn later) */
    for (check = size - 2; check >= 0; check--) {
	if (table[check].objptr->flags & CHECKPOLY) {
	    for (tablepos = size - 1; tablepos > check; tablepos--) {
		if (table[tablepos].objptr->flags & CHECKOBJ) {
		    switch (table[check].objptr->flags & CHECKPOLY) {
			case CHECKX :
			    planecoord = table[check].displ.x;
			    objcoord = table[tablepos].displ.x;
			    viewercoord = viewer.position.x;
			    break;
			case CHECKY :
			    planecoord = table[check].displ.y;
			    objcoord = table[tablepos].displ.y;
			    viewercoord = viewer.position.y;
			    break;
			case CHECKZ :
			    planecoord = table[check].displ.z;
			    objcoord = table[tablepos].displ.z;
			    viewercoord = viewer.position.z;
			    break;
			default :
			    fprintf(stderr,
			      "Illegal CHECKPOLY flag combination.\n");
			    exit(1);
		    }
		    /* Plane needs to be moved towards end of table if
		     object and viewer are on different sides of the plane */
		    if (sign(objcoord - planecoord) !=
		      sign(viewercoord - planecoord)) {
			int i;
			point *points, temp;
			long  maxx, maxy, maxz;

			maxx = maxy = maxz = 0L;
			points = table[check].objptr->points;
			for (i = 0; i < table[check].objptr->no_points; i++) {
			    temp = points[i];
			    rotpoint(&table[check].objptr->rotmatrix,
				     &temp);
			    if (labs(temp.x) > maxx)
				maxx = labs(temp.x);
			    if (labs(temp.y) > maxy)
				maxy = labs(temp.y);
			    if (labs(temp.z) > maxz)
				maxz = labs(temp.z);
			}

			/* If the object being checked is outside the area
			   covered by the plane object then continue
			   inner loop */
			if (table[check].objptr->flags & CHECKX) {
			    if (labs(table[tablepos].displ.y -
				     table[check].displ.y) > maxy) continue;
			    if (labs(table[tablepos].displ.z -
				     table[check].displ.z) > maxz) continue;
			}
			else if (table[check].objptr->flags & CHECKY) {
			    if (labs(table[tablepos].displ.x -
				     table[check].displ.x) > maxx) continue;
			    if (labs(table[tablepos].displ.z -
				     table[check].displ.z) > maxz) continue;
			}
			else {	/* CHECKZ must be set */
			    if (labs(table[tablepos].displ.x -
				     table[check].displ.x) > maxx) continue;
			    if (labs(table[tablepos].displ.y -
				     table[check].displ.y) > maxy) continue;
			}

			tempentry = table[check];
			bcopy((char *)&table[check + 1],
				(char *)&table[check],
				(tablepos - check) * sizeof(sortentry));
			table[tablepos] = tempentry;
			break;
		    }
		}
	    }
	}
    }
}  /* depth_sort() */


/* Draw all objects */
void draw_all(
	disp *d,
	sortentry *table,
	int size,
	long x,
	long y,
	long sc,
	object **LOS_obj,    /* Object which is in line of sight */
	char **clxdata,      /* Pass back collidedata pointer */
	char **portal,	     /* Pass back portaldata pointer */
	point *teleport)
{
    int i;
    static point rotpoints[MAXPOINTS];
    static screenpoint screenpoints[MAXPOINTS];
    long LOS_z;  /* Z coordinate of object in line of sight */

    scale = sc;
    centrex = x;
    centrey = y;
    displ = d;
    collidedata = NULL;
    teleportdata = *teleport;
    portaldata = NULL;
    LOS_z = -1L;  /* No object in LOS */
    y_collisions = z_collisions = 0;
    original_x = viewer.position.x;
    original_z = viewer.position.z;

    if (size) {
	if (size > 1) depth_sort(table, size);
	for (i = 0; i < size; i++) {
	    rotate_points(&table[i], rotpoints);
	    do_perspective(rotpoints, screenpoints,
			   table[i].objptr->no_points);
	    draw_object(rotpoints, screenpoints,
			 table[i].objptr, &LOS_z, LOS_obj);
	}
    }
    if (y_collisions || z_collisions) {
	viewer.yvel = 0L;
	viewer.zvel = 0L;
    }
    if (LOS_z < 0) *LOS_obj = NULL;  /* No object in LOS */
    *clxdata = collidedata;
    *portal = portaldata;
    *teleport = teleportdata;
}  /* draw_all() */

