/*
 * client.c in Virtual Combat
 * Copyright (c) 1993 by David Jameson
 *
 * Permission to use, copy, modify and distribute this software
 * for any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission appear in the supporting
 * documentation.
 */

/* Client program for multi-user 3D environment */

#include "defs.h"

#include <netdb.h>
#include <math.h>
#include <signal.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <limits.h>
#include <errno.h>
#include <assert.h>


#define GRAVITY 80  /* Acceleration due to gravity (mm/frame/frame) */

/*
 * SPEED is the value of player movement per turn for when the game is
 * running at 10 frames per second.  If the game is running faster or
 * slower than this speed then the value is scaled appropriately.
 */

#define SPEED 400	     /* Speed of movement for player (mm/frame) */
#define MAXSPEED 600	     /* Maximum speed */
#define BLAST_DISTANCE 2000  /* Blast distance of grenades */
#define PLAYER_CLXRADIUS 1500	/* Collision distance with other players */
#define MIN_DIST 4000	/* Minimum distance for shooting/grenades to work */
#define THETA (25.0 * M_PI / 180.0)   /* Angle that grenades are launched at */
#define MAXANGLE (40.0 * M_PI / 180)  /* Maximum value for xrot */
#define UP 1
#define DOWN 2
#define ROTDIV 180   /* Sensitivity of control using mouse (higher = less
			sensitive */
#define MAX_Y 100    /* If you fall below this you're dead */
#define MAX_TEMP 50  /* Max laser temperature */
#define MAX_GRENADES 3
#define VIEWING_ANGLE (90.0 * M_PI / 180.0)

#include "3d.h"
#include "protocol.h"


typedef struct {
    unsigned char q, a, o, p;
} keymap;


/* Module variables */

static disp *displ;   /* Display information */
static char *dpyname;
static int start_position;   /* Order of starting (0 = 1st to join game) */
static long xsize, ysize, centrex, centrey;
static int frames;
static int framespersec;
static int grenade;  /* Object number of grenade (or -1) */
static int sckt;
static char *name1 = "Virtual Combat";
static char *name2 = "vircom";
static char version[] = "qub_1.0";  /* Version of game */
static XColor red, black, skycol, skycol2, blue;
static int ok_to_send;
static point grenade_pos, grenade_vel;
static int no_grenades;        /* No. of grenades player has */
static databuffer inbuffer;    /* Buffer for incoming data */
static databuffer textbuffer;  /* Buffer for outgoing text */
static databuffer gamebuffer;  /* Buffer for outgoing game data */
static int id;	/* Each client receives a unique ID when they connect */
static point teleport;	/* New position if player collides with portal */
static int mode;			   /* TEXTMODE, GAMEMODE etc. */
static int warp;	   /* Warp pointer? */
static int energy;	   /* Player's energy */
static int temp;	   /* Laser temperature */
static int portal_collision;
static Cursor cross;
static long scale;
static object *LOS_obj;    /* Object in line of sight */
static int lastenergy;	   /* Previous energy */
static char *collidedata;  /* Message sent to server when a collision
			      occurs */
static char *portaldata;   /* Message sent to server when a collision
			      occurs with a portal object */
static object *world;	   /* Pointer to "the world" i.e. root object of
			      everything which is drawn on screen
			      (except other players) */
static point start[MAX_PLAYERS];  /* Possible start locations */
static int numstart;		  /* Size of start[] array */
viewer_object viewer;	          /* Information about the viewer */

/*
 * Each other player in the game has a 3d object which represents the
 * player in the game.	A pointer to the object is stores in
 * playerobj[id] (0 < id < MAX_LOGINS)
 * Pointers to grenade objects are also stored in the playerobj array,
 * the grenade thrown by player id being stored in position
 * MAX_LOGINS + id
 */

static object *playerobj[MAX_LOGINS * 2];	/* Other players in game */


/* Prototypes */

static void usage(char *name);
static void sighandler(int sig);
static void alrm(int sig);
static void senddata(int socket, unsigned char *data, int length);
static void draw_sight(void);
static void draw_sight2(void);
static void draw_grenades(void);
static void set_size(long x, long y);
static void do_player_clx(void);
static void draw_laser(void);
static void draw_temp_bar(void);
static void draw_energy_bar(void);
static void get_input(void);
static int update_position(keymap *map);
static void do_events(void);
static void launch_grenade(void);
static int find_space(void);
static void leave_gamemode(void);
static void send_collision(char *data);
static void send_collision_portal(char *data);
static void send_shoot(char *data);
static void send_data_to_server(void);
static void get_mouse(int *x, int *y, unsigned int *mask);
static void connect_to_server(char *name);
static void send_explosion(int id);
static void parse_all(int sock, unsigned char *text);
static void sendtxt(int sock, unsigned char *text);
static void recvdata(int socket, unsigned char *data, int length);



/* senddata() and recvdata() are versions of write() and read() which
   are guaranteed to transfer the length of data requested */

static void senddata(int socket, unsigned char *data, int length)
{
    int sent;

    while (length > 0) {
	sent = write(socket, data, length);
	if (sent == 0) {
	    close(socket);
	    exit(0);
	}
	if (sent == -1) {
	    perror("write()");
	    close(socket);
	    exit(1);
	}
	length -= sent;
	data += sent;
#ifdef DEBUG
	if (length) fprintf(stderr, "write(): still %d bytes to write\n",
			    length);
#endif
    }
}  /* senddata() */


static void recvdata(int socket, unsigned char *data, int length)
{
    int got;

    while (length > 0) {
	got = read(socket, data, length);
	if (got == 0) {
	    close(socket);
	    exit(0);
	}
	if (got == -1) {
	    perror("read()");
	    close(socket);
	    exit(1);
	}
	length -= got;
	data += got;
#ifdef DEBUG
	if (length) fprintf(stderr, "read(): still %d bytes to read\n",
			    length);
#endif
    }
}  /* recvdata() */


static void sendtxt(int sock, unsigned char *text)
{
    int slen;

    slen = strlen(text);
    if (slen < DATASIZE) strcpy(textbuffer.data, text);
    else {
	memcpy(textbuffer.data, text, DATASIZE - 1);
	textbuffer.data[DATASIZE - 1] = 0;
	slen = DATASIZE - 1;
    }
    textbuffer.length = htons(slen + 1 + HDRSIZE);
    textbuffer.pkttype = TEXTDATA;
    senddata(sock, (unsigned char *)&textbuffer, ntohs(textbuffer.length));
}  /* sendtxt() */


/* Function called when user presses ctrl-C or kills the client process */
static void sighandler(int sig)
{
    if (mode & GAMEMODE) leave_gamemode();
    close(sckt);
    printf("Exiting program.\n");
    if (mode & GAMEMODE) {
	XAutoRepeatOn(displ->dpy);
	XSync(displ->dpy, False);
	XCloseDisplay(displ->dpy);
    }
    exit(0);
}  /* sighandler() */


/* Called once every second */
static void alrm(int sig)
{
    Window win;
    int revert;

    /* First check to make sure that game window still has input focus */

    /*if ((mode & GAMEMODE) && warp) {
	XGetInputFocus(displ->dpy, &win, &revert);
	if (win != displ->w) warp = FALSE;
    }*/

    framespersec = frames;  /* Speed of game in frames per second */
    frames = 0;
}  /* alrm() */


static void parse_all(int sock, unsigned char *text)
{
    int slen;

    slen = strlen(text);
    if (slen < DATASIZE) strcpy(textbuffer.data, text);
    else {
	memcpy(textbuffer.data, text, DATASIZE - 1);
	textbuffer.data[DATASIZE - 1] = 0;
	slen = DATASIZE - 1;
    }
    textbuffer.length = htons(slen + 1 + HDRSIZE);
    textbuffer.pkttype = GAMEPARSE;
    senddata(sock, (unsigned char *)&textbuffer, ntohs(textbuffer.length));
}  /* parse_all() */


/* Try to initiate connection to server */
static void connect_to_server(char *name)
{
    struct sockaddr_in sin1;
    struct hostent *host_entry;
    struct in_addr address;
    sckt = socket(AF_INET, SOCK_STREAM, 0);
    if (sckt < 0) {
	perror("socket()");
	exit(1);
    }
    if (isdigit(name[0])) {     /* Numeric IP address */
	if ((address.s_addr = inet_addr(name)) == 0xffffffff) {
	    perror("Error resolving IP address");
	    exit(1);
	}
    }
    else {	/* DNS host name */
	if ((host_entry = gethostbyname(name)) == NULL) {
	    perror("Error resolving host name");
	    exit(1);
	}
	bcopy(host_entry->h_addr, (char *)&address, sizeof(struct in_addr));
    }
    sin1.sin_addr.s_addr = address.s_addr;
    sin1.sin_family = AF_INET;
    sin1.sin_port = htons((u_short)PORT_NUMBER);
    if (connect(sckt, (struct sockaddr *)&sin1, sizeof sin1) < 0) {
	perror("connect()");
	exit(1);
    }
    textbuffer.length = htons(HDRSIZE + 1);
    sendtxt(sckt, version);
    recvdata(sckt, (unsigned char *)&inbuffer, HDRSIZE + 1);
    if (inbuffer.pkttype == TEXTNOLOGIN) {
	printf("Too many players logged onto server.\n");
	close(sckt);
	exit(0);
    }
    else if (inbuffer.pkttype != TEXTLOGIN) {
	fprintf(stderr, "Error connecting to server.\n");
	close(sckt);
	exit(1);
    }

    id = inbuffer.data[0];
}  /* connect_to_server() */


/* Tell server that our grenade has exploded */
static void send_explosion(int id)
{
    gamebuffer.data[0] = id;
    gamebuffer.length = htons(HDRSIZE + 1);
    gamebuffer.pkttype = EXPLOSION;
    senddata(sckt, (unsigned char *)&gamebuffer, ntohs(gamebuffer.length));
}  /* send_explosion() */


/* Get change in mouse position since last time this function was called */
static void get_mouse(int *x, int *y,
  unsigned int *mask)  /* Mask returning which mouse buttons are pressed */
{
    static Window root_return, child_return;
    int root_x, root_y;
    int win_x, win_y;
    Window win;
    int revert;

    XQueryPointer(displ->dpy, displ->w, &root_return, &child_return, &root_x,
		  &root_y, &win_x, &win_y, mask);
    if (warp && win_x != centrex)
	XWarpPointer(displ->dpy, None, None, 0, 0, 0, 0, centrex - win_x, 0);
    if (warp && win_y != centrey)
	XWarpPointer(displ->dpy, None, None, 0, 0, 0, 0, 0, centrey - win_y);
	/*XSync (displ->dpy, 0);*/
    if ((! warp) && ((*mask & Button1Mask) || (*mask & Button2Mask)) &&
      (win_x < xsize - 5) && (win_y < ysize - 5) && (win_x > 5) &&
      (win_y > 25)) {
	XGetInputFocus(displ->dpy, &win, &revert);
	if (win == displ->w) {
	    warp = TRUE;
	    XWarpPointer(displ->dpy, None, displ->w, 0, 0, 0, 0, centrex, centrey);
	/*XSetInputFocus(displ->dpy, displ->w, RevertToPointerRoot, CurrentTime);*/
	/*XGrabPointer(displ->dpy, displ->w, False, PointerMotionMask,
		     GrabModeAsync, GrabModeAsync,
		     displ->w, None, CurrentTime);*/
	}
    }
    *x = win_x - centrex;
    *y = win_y - centrey;
}  /* get_mouse() */


/* Send position and orientation to server */
static void send_data_to_server(void)
{
    update upd;
    char buf[100];
    int i;

    if (grenade >= 0) {
	for (i = 0; i < MAX_LOGINS; i++) {
	    if (playerobj[i]) {
		if (distance(&playerobj[i]->centre, &grenade_pos) <
		  BLAST_DISTANCE) {
		    send_explosion(i);
		    sprintf(buf, "remove \"grenade_%d\";", id);
		    parse_all(sckt, buf);
		    grenade = -1;
		    break;
		}
	    }
	}
	if (grenade_pos.y > MAX_Y) {
	    sprintf(buf, "remove \"grenade_%d\";", id);
	    parse_all(sckt, buf);
	    grenade = -1;
	}
    }
    if (grenade >= 0) {
	/* Send updated position to server */
	grenade_pos.x += grenade_vel.x;
	grenade_pos.y += grenade_vel.y;
	grenade_pos.z += grenade_vel.z;
	grenade_vel.y += GRAVITY;

#if DEBUG >= 3
	fprintf (stderr, "Grenade position: %ld, %ld, %ld\n", grenade_pos.x,
	    grenade_pos.y, grenade_pos.z);
#endif

	upd.x = htonl(grenade_pos.x);
	upd.y = htonl(grenade_pos.y);
	upd.z = htonl(grenade_pos.z);
	upd.yrot = htonl(0L);
	upd.id = grenade;

	gamebuffer.length = htons(HDRSIZE + sizeof(update));
	gamebuffer.pkttype = GAMEDATA;
	memcpy(gamebuffer.data, &upd, sizeof(update));
	senddata(sckt, (unsigned char *)&gamebuffer,
	         ntohs(gamebuffer.length));
    }

    if (portal_collision) {
	/* We haven't been transported to the new world yet */
	upd.x = htonl(100000);
	upd.y = htonl(-100000);
	upd.z = htonl(100000);
    }
    else {
	upd.x = htonl(viewer.position.x);
	upd.y = htonl(viewer.position.y);
	upd.z = htonl(viewer.position.z);
    }
    upd.yrot = htonl((long)(viewer.yrot * ROTDIV));
    upd.id = id;

    gamebuffer.length = htons(HDRSIZE + sizeof(update));
    gamebuffer.pkttype = GAMEDATA;
    memcpy(gamebuffer.data, &upd, sizeof(update));
    senddata(sckt, (unsigned char *)&gamebuffer, ntohs(gamebuffer.length));
}  /* send_data_to_server */


/* Tell server that we have collided with something */
static void send_collision(
	char *data)  /* text string identifying the thing that we've hit */
{
    strcpy(gamebuffer.data, data);
    gamebuffer.length = htons(HDRSIZE + strlen(data) + 1);
    gamebuffer.pkttype = GAMECOLLIDE;
    senddata(sckt, (unsigned char *)&gamebuffer, ntohs(gamebuffer.length));
}  /* send_collision() */


/* Tell server that we have collided with a portal */
static void send_collision_portal(
	char *data)  /* text string identifying filename of new world */
{
    char buf[100];

    if (grenade >= 0)
    {
	sprintf(buf, "remove \"grenade_%d\";", id);
	parse_all(sckt, buf);
	grenade = -1;
    }

    strcpy(gamebuffer.data, data);
    gamebuffer.length = htons(HDRSIZE + strlen(data) + 1);
    gamebuffer.pkttype = GAMEPORTAL;
    senddata(sckt, (unsigned char *)&gamebuffer, ntohs(gamebuffer.length));
}  /* send_collision_portal() */


/* Tell server that we have shot something */
static void send_shoot(
	 char *data)  /* text string identifying the thing that we've shot */
{
    strcpy(gamebuffer.data, data);
    gamebuffer.length = htons(HDRSIZE + strlen(data) + 1);
    gamebuffer.pkttype = GAMESHOOT;
    senddata(sckt, (unsigned char *)&gamebuffer, ntohs(gamebuffer.length));
}  /* send_shoot() */


/* Called when player leaves game mode */
static void leave_gamemode(void)
{
    int i;
    char buf[100];

    if (grenade >= 0)
    {
	sprintf(buf, "remove \"grenade_%d\";", id);
	parse_all(sckt, buf);
	grenade = -1;
    }

    mode ^= GAMEMODE;

    /* Free memory allocated for objects */

    destroy(world);
    for (i = 0; i < MAX_LOGINS * 2; i++)
	if (playerobj[i]) destroy(playerobj[i]);
}  /* leave_gamemode() */


/* Find space in playerdata[] table for a moving object */
static int find_space(void)
{
    return id + MAX_LOGINS;
}  /* find_space() */


/* Launch a grenade */
static void launch_grenade(void)
{
    char buf[100];
    float height;  /* Height of target above viewer */
    float dist;    /* Horizontal distance of target from viewer */
    point temp;
    float speed;   /* Initial speed of projectile */
    float pos;
    object *aimobj;  /* Object we are aiming at */

    if (! no_grenades) return;

    grenade_pos = viewer.position;

    /*
     * Find the parent object of the thing being aimed at, so that we
     * have the absolute position of the object instead of the position
     * relative to the parent object.
     */

    aimobj = LOS_obj;
    while (aimobj->parentobj != world && aimobj->parentobj)
	aimobj = aimobj->parentobj;

    height = viewer.position.y - aimobj->centre.y;
    temp = aimobj->centre;

    /*
     * To find the horizontal distance between the viewer and target, set
     * the y-coords equal and calculate the distance between them.
     */

    temp.y = viewer.position.y;
    dist = distance(&viewer.position, &temp);

    pos = dist * tan(THETA) - height;
    if (pos > 0.0001) {
	--no_grenades;
	grenade = find_space();
#if DEBUG >= 2
	fprintf (stderr, "Grenade launched, ID=%d\n", grenade);
#endif
	sprintf(buf, "player \"grenade_%d\" %d \"grenade\";", id, grenade);
	parse_all(sckt, buf);

	speed = sqrt(GRAVITY / 2 / pos) * dist / cos(THETA);
	polar_to_xyz (&grenade_vel, speed, -THETA, viewer.yrot);
#if DEBUG >= 2
	fprintf(stderr, "grenade speed: %f, velocity: %ld, %ld, %ld\n",
		speed, grenade_vel.x, grenade_vel.y, grenade_vel.z);
#endif
	}
}  /* launch_grenade() */


static void do_events(void)
{
    XEvent ev;
    KeySym ks;
    XResizeRequestEvent *resize;
    char str[256+1];
    int bytes;
    object *aimobj;
    int i;

    while (XPending(displ->dpy)) {
	XNextEvent(displ->dpy, &ev);
	switch(ev.type) {
	    case KeyPress :
		bytes = XLookupString(&ev.xkey, str, 256, &ks, NULL);
		str[bytes] = '\0';
		if (ks != NoSymbol) {
		    if (str[0] == ' ') warp = !warp;
		    else if (str[0] == 'f')
			printf(
    "Game running at %d frames per second.\n", framespersec);
		    else if (str[0] == 'x') {
			leave_gamemode();
			gamebuffer.length = htons(HDRSIZE);
			gamebuffer.pkttype = LEAVEGAME;
			senddata(sckt, (unsigned char *)&gamebuffer,
				 ntohs(gamebuffer.length));
		    }
		    else if (str[0] == 'i') {
			if (LOS_obj && LOS_obj->shoot) {
			    aimobj = LOS_obj;
			    while (aimobj->parentobj != world &&
				   aimobj->parentobj)
				aimobj = aimobj->parentobj;
			    printf("Player \"%s\"\n", aimobj->shoot);
			}
		    }
		    else if (str[0] == 'w') {
			/*
			 * This key originally displayed the players who
			 * were in the same room, but this gave players
			 * an unfair advantage.
			 */
			/*puts("Other players in this room:");
			for (i = 0; i < MAX_LOGINS; i++) {
			    if (playerobj[i]) puts(playerobj[i]->shoot);
			}*/
			gamebuffer.length = htons(HDRSIZE);
			gamebuffer.pkttype = GAMEWHO;
			senddata(sckt, (unsigned char *)&gamebuffer,
				 ntohs(gamebuffer.length));
		    }
		    else if (ks == XK_Return) {
			if (LOS_obj && LOS_obj->shoot) {
			    aimobj = LOS_obj;
			    while (aimobj->parentobj != world &&
				   aimobj->parentobj)
				aimobj = aimobj->parentobj;
			    if (grenade < 0 && distance(&aimobj->centre,
				    &viewer.position) > MIN_DIST)
				launch_grenade();
			}
		    }
#ifdef DEBUG
		    else printf("Key %s not defined.\n", str);
#endif
		}
		break;
	    case ResizeRequest :
		resize = (XResizeRequestEvent *)&ev;
		set_size(resize->width, resize->height);
		break;
	    case KeyRelease :
		break;
	    case ButtonPress :
		break;
	    case ButtonRelease :
		break;
	    case Expose :
		if (ev.xexpose.count == 0) {
		    Window dummyw;
		    int dummyi;
		    unsigned int dummyu, width, height;

		    XGetGeometry(displ->dpy, displ->w, &dummyw, &dummyi,
				 &dummyi, &width, &height, &dummyu, &dummyu);
		    set_size(width, height);
		}
		break;
	    case FocusIn :
		break;
	    case FocusOut :
		warp = FALSE;
		break;
	    case ClientMessage :
		break;
	    default :
#ifdef DEBUG
		fprintf(stderr, "Unexpected X-event: %d\n", ev.type);
#endif
		break;
	}
    }
}  /* do_events() */

#if 0
/* Routine to process a key being pressed */
static void do_keypress(event, map)
XEvent *event;
keymap *map;
{
    XKeyEvent *keyev = (XKeyEvent *)event;
    KeySym ks;
    char str[256+1];
    int bytes;
    char buf[100];
    int i;

    bytes = XLookupString(keyev, str, 256, &ks, NULL);
    if (ks != NoSymbol) {
	if (str[0] == ' ') warp = !warp;
	if (str[0] == 'u') warp = FALSE;
	if (str[0] == 'w') warp = TRUE;
	if (str[0] == 'x' || str[0] == 'q') {
	    leave_gamemode();
	    gamebuffer.length = htons(HDRSIZE);
	    gamebuffer.pkttype = LEAVEGAME;
	    senddata(sckt, (unsigned char *)&gamebuffer,
	             ntohs(gamebuffer.length));
	}
	if (str[0] == '\n' && (grenade < 0)) {
	    grenade = find_space();
#if DEBUG >= 2
	    fprintf (stderr, "Grenade launched, ID=%d\n", grenade);
#endif
	    sprintf(buf, "player \"grenade_%d\" %d \"grenade\";",
		    id, grenade);
	    parse_all(sckt, buf);
	}
    }
}  /* do_keypress */


/* Routine to process a key being released */
static void do_keyrelease(event, map)
XEvent *event;
keymap *map;
{
    XKeyEvent *keyev = (XKeyEvent *)event;
    KeySym ks;
    char str[256+1];
    int bytes;

    bytes = XLookupString(keyev, str, 256, &ks, NULL);
    if (ks != NoSymbol) {
    }
}  /* do_keyrelease */
#endif


/* Check for key presses/releases and update viewer position/orientation */
static int update_position(keymap *map)
{
    matrix *mat = &viewer.rotmatrix;
    int dx, dy;  /* Change in mouse position */
    unsigned int mask;

#if DEBUG >= 3
    fprintf (stderr, "Position: %ld, %ld, %ld\n", viewer.position.x,
	viewer.position.y, viewer.position.z);
#endif
    if (viewer.position.y > MAX_Y) {
	leave_gamemode();
	gamebuffer.length = htons(HDRSIZE);
	gamebuffer.pkttype = LEAVEGAME;
	senddata(sckt, (unsigned char *)&gamebuffer,
	         ntohs(gamebuffer.length));
	printf("You fell off the edge of the world!\n");
	return 0;
    }

    do_events();

    get_mouse(&dx, &dy, &mask);
    if (! warp) return 0;

    /* Check velocity  */

    if (! viewer.yvel) {  /* If player is falling, don't change z velocity */
	viewer.zvel = 0L;
	if (mask & Button1Mask) {
	    if (framespersec) viewer.zvel = (SPEED * 10) / framespersec;
	    else viewer.zvel = MAXSPEED;
	    if (viewer.zvel > MAXSPEED) viewer.zvel = MAXSPEED;
	}
	if (mask & Button2Mask) {
	    if (framespersec) viewer.zvel = (- SPEED * 10) / framespersec;
	    else viewer.zvel = -MAXSPEED;
	    if (viewer.zvel < -MAXSPEED) viewer.zvel = -MAXSPEED;
	}
    }

    /* Check rotation */
    if (dx) {
	viewer.yrot += (float)dx / ROTDIV;
	if (viewer.yrot > 2 * M_PI) viewer.yrot -= 2 * M_PI;
	if (viewer.yrot < 0) viewer.yrot += 2 * M_PI;
    }
    if (dy) {
	viewer.xrot -= (float)dy / ROTDIV;
	if (viewer.xrot < 0) viewer.xrot += 2 * M_PI;
	if (viewer.xrot > 2 * M_PI) viewer.xrot -= 2 * M_PI;
	if ((viewer.xrot < (2 * M_PI - MAXANGLE)) &&
	   (viewer.xrot > M_PI)) viewer.xrot = 2 * M_PI - MAXANGLE;
	if ((viewer.xrot > MAXANGLE) && (viewer.xrot < M_PI))
	   viewer.xrot = MAXANGLE;
    }

    /* If viewer is moving then set xrot to 0 (necessary for collision
       detection) */
    if (viewer.zvel || viewer.yvel) viewer.xrot = 0.0;

    /* Update rotation matrices */
    unit_matrix(&viewer.rotmatrix);
    viewer.inversematrix = viewer.rotmatrix;
    rotate_x(&viewer.rotmatrix, viewer.xrot, WORLDMODE);
    rotate_y(&viewer.rotmatrix, viewer.yrot, WORLDMODE);

    rotate_y(&viewer.inversematrix, -viewer.yrot, WORLDMODE);
    rotate_x(&viewer.inversematrix, -viewer.xrot, WORLDMODE);

    /* Now  position using velocity and viewer's rotation matrix.
       Velocity only has a z component before it is rotated. */

    if (viewer.zvel) viewer.yvel += GRAVITY;
    if (viewer.yvel > MAXSPEED) viewer.yvel = MAXSPEED;
    viewer.position.y += viewer.yvel;

    viewer.position.x += mat->a3 * viewer.zvel;
    /*viewer.position.y += mat->b3 * viewer.zvel;*/  /* Will be zero */
    viewer.position.z += mat->c3 * viewer.zvel;
    return mask & Button3Mask;
}  /* update_position */


/*
 * Routine to do collision detection between this player and other players.
 * Collision detection between players and objects is done by a separate
 * routine (in 3d.c)
 */ 
static void do_player_clx(void)
{
    int i;
    point player_pos;
    float dist, displ_x, displ_z;

    for (i = 0; i < MAX_LOGINS; i++) {
	if (playerobj[i]) {
	    player_pos = playerobj[i]->centre;
	    displ_x = viewer.position.x - player_pos.x;
	    displ_z = viewer.position.z - player_pos.z;
	    dist = sqrt((double)(displ_x * displ_x + displ_z * displ_z));
	    if ((dist < PLAYER_CLXRADIUS) && (viewer.position.y - player_pos.y
		< CLXHEIGHT) && (viewer.position.y - player_pos.y >
		 -CLXHEIGHT)) {
		displ_x *= PLAYER_CLXRADIUS;
		if (dist != 0.0) displ_x /= dist;

		displ_z *= PLAYER_CLXRADIUS;
		if (dist != 0.0) displ_z /= dist;

		viewer.position.x = player_pos.x + displ_x;
		viewer.position.z = player_pos.z + displ_z;
	    }
	}
    }
}  /* do_player_clx() */


/* Wait for input from either keyboard or socket */
static void get_input(void)
{
    unsigned char *ptr;
    update upd;
    int oldmode;
    int nds;  /* Max number of file descriptors (for select() ) */
    fd_set readfds;
    int num;
    struct timeval timeout;
    point rel;
    point player_pos;

    nds = getdtablesize();
    oldmode = mode;

    do {
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	if (! (mode & GAMEMODE)) timeout.tv_sec = 10000;
	FD_ZERO(&readfds);
	FD_SET(fileno(stdin), &readfds);
	FD_SET(sckt, &readfds);
	if ((num = select(nds, &readfds, NULL, NULL, &timeout)) > 0) {
	    if (FD_ISSET(fileno(stdin), &readfds) && (mode & TEXTMODE)) {
		if (fgets(inbuffer.data, DATASIZE, stdin)) {
		    int slen;

		    slen = strlen(inbuffer.data);
		    if ((slen > 0) && (inbuffer.data[slen - 1] == '\n'))
			inbuffer.data[slen - 1] = '\0';
		    sendtxt(sckt, inbuffer.data);
		}
	    }
	    else if (FD_ISSET(sckt, &readfds)) {
		recvdata(sckt, (unsigned char *)&inbuffer,
			 sizeof inbuffer.length);
		recvdata(sckt, (unsigned char *)&inbuffer +
		  sizeof inbuffer.length,
		  ntohs(inbuffer.length) - sizeof inbuffer.length);
#if DEBUG >= 3
		fprintf(stderr, "Received packet %u, length %u\n",
			inbuffer.pkttype, ntohs(inbuffer.length));
#endif
		if ((mode & TEXTMODE) && (inbuffer.pkttype & TEXTPACKET))
		  switch (inbuffer.pkttype) {
		    case TEXTDATA :
			if (ntohs(inbuffer.length) > HDRSIZE) {
			    fputs(inbuffer.data, stdout);
			    fflush(stdout);
			}
			break;
		    case TEXTECHOON :
			local_echo_on();
			break;
		    case TEXTECHOOFF :
			local_echo_off();
			break;
		    case ENTERGAME :
			start_position = inbuffer.data[0];
			mode |= GAMEMODE;
			return;
		    case TEXTCLOSE :
			mode = CLOSEDOWN;
			break;
		}
		else if ((mode & GAMEMODE) &&
		  (inbuffer.pkttype & GAMEPACKET)) {
		    switch (inbuffer.pkttype) {
			case GAMEDATA :
			    ok_to_send = TRUE;
			    ptr = inbuffer.data;
			    while (ptr < (inbuffer.data +
			      ntohs(inbuffer.length) - HDRSIZE)) {
				memcpy((char *)&upd, ptr, sizeof(update));
#if DEBUG >= 3
			fprintf(stderr, "\tGamedata received from: %d\n",
				upd.id);
#endif
				if (upd.id != id) {
				    if (playerobj[upd.id]) {
					player_pos.x = ntohl(upd.x);
					player_pos.y = ntohl(upd.y);
					player_pos.z = ntohl(upd.z);
					playerobj[upd.id]->centre = player_pos;
					unit_matrix(&playerobj[upd.id]->rotmatrix);
					rotate_y(&playerobj[upd.id]->rotmatrix,
		      (float)ntohl(upd.yrot) / ROTDIV, WORLDMODE);
				    }
				}
				ptr += sizeof(update);
			    }
			    break;
			case LEAVEGAME :
			    leave_gamemode();
			    return;
			case GAMECLOSE :
			    leave_gamemode();
			    mode = CLOSEDOWN;
			    break;
			case GAMEPARSE :
			    parse(displ, inbuffer.data,
				&start[MAX_PLAYERS - numstart], &numstart,
				&world, playerobj);
			    if (portal_collision) {
				viewer.position = teleport;
				viewer.position.y -= CLXHEIGHT;
				portal_collision = FALSE;
			    }
			    else if (MAX_PLAYERS - numstart > start_position) {
				viewer.position = start[start_position];
				viewer.position.y -= CLXHEIGHT;
				numstart = MAX_PLAYERS;
			    }
			    break;
			case ENERGYLOSS :
			    energy -= inbuffer.data[0];
			    break;
			case ENERGYGAIN :
			    energy += inbuffer.data[0];
			    break;
		    }
		}
	    }
	}
	if (num < 0 && errno != EINTR) {
	    perror("select()");
	    exit(1);
	}
    } while (num > 0);
}  /* get_input() */


static void draw_energy_bar(void)
{
    XSetForeground(displ->dpy, displ->gc, black.pixel);
    XDrawRectangle(displ->dpy, (Drawable)displ->pix, displ->gc, 50, ysize - 80,
		   100, 20);
    XSetForeground(displ->dpy, displ->gc, blue.pixel);
    XFillRectangle(displ->dpy, (Drawable)displ->pix, displ->gc, 54, ysize - 76,
		   (92 * energy) / MAX_ENERGY, 12);
}  /* draw_energy_bar() */


static void draw_temp_bar(void)
{
    XSetForeground(displ->dpy, displ->gc, black.pixel);
    XDrawRectangle(displ->dpy, (Drawable)displ->pix, displ->gc, 50, ysize - 50,
		   100, 20);
    XSetForeground(displ->dpy, displ->gc, red.pixel);
    XFillRectangle(displ->dpy, (Drawable)displ->pix, displ->gc, 54, ysize - 46,
		   (92 * temp) / MAX_TEMP, 12);
}  /* draw_temp_bar() */


static void draw_grenades(void)
{
    int i;

    XSetForeground(displ->dpy, displ->gc, black.pixel);
    for (i = 0; i < no_grenades; i++)
	XFillRectangle(displ->dpy, (Drawable)displ->pix, displ->gc,
		       xsize - 100 + i * 25, ysize - 76, 20, 20);
}  /* draw_grenades() */


static void draw_sight(void)
{
    XSetForeground(displ->dpy, displ->gc, red.pixel);
    XDrawRectangle(displ->dpy, (Drawable)displ->pix, displ->gc,
		       centrex - 20, centrey - 20, 40, 40);
}  /* draw_sight() */


/* Drawn when an enemy is too close to shoot */
static void draw_sight2(void)
{
    XSetForeground(displ->dpy, displ->gc, black.pixel);
    XDrawRectangle(displ->dpy, (Drawable)displ->pix, displ->gc,
		       centrex - 20, centrey - 20, 40, 40);
}  /* draw_sight2() */


static void draw_laser(void)
{
    int randx, randy;

    randx = (10.0 * (rand() % 256)) / 256 - 5.0;
    randy = (10.0 * (rand() % 256)) / 256 - 5.0;
    XSetForeground(displ->dpy, displ->gc, red.pixel);
    XDrawLine(displ->dpy, (Drawable)displ->pix, displ->gc,
	      centrex + randx, centrey + randy, 0, ysize);
    XDrawLine(displ->dpy, (Drawable)displ->pix, displ->gc,
	      centrex + randx, centrey + randy, xsize, ysize);
}  /* draw_laser() */


static void usage(char *name)
{
	fprintf(stderr,
	    "Usage: %s [<address>] [-display <display name>]\n", name);
}  /* usage() */


/* Set the size of the game window */
static void set_size(long x, long y)
{
    xsize = x;
    ysize = y;
    centrex = xsize / 2;
    centrey = ysize / 2;
    scale = (float)xsize / tan(VIEWING_ANGLE / 2.0);
    XFreePixmap(displ->dpy, displ->pix);
    displ->pix = XCreatePixmap(displ->dpy, (Drawable)displ->w, x, y, PLANES);
}



int main(int argc, char *argv[])
{
    keymap map;
    sortentry table[TABLESIZE];
    char hostfilename[200];
    int i;
    int trigger;  /* Is player shooting? */
    char *hostname;
    float dist;  /* Distance to object in line of sight */
    char hostbuf[101];
    FILE *hostfile;
    object *aimobj;
    struct itimerval itime, itime2;


    hostname = dpyname = NULL;
    portal_collision = FALSE;
    mode = TEXTMODE;


    /* Set up a signal handler to catch ctrl-C and 'kill's */

    signal(SIGINT, sighandler);
    signal(SIGTERM, sighandler);

    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-') {
	    switch (argv[i][1]) {
		case 'd' :
		    if (++i >= argc) {
			usage(argv[0]);
			exit(1);
		    }
		    dpyname = argv[i];
		    break;
		default :
		    usage(argv[0]);
		    exit(1);
		    break;
	    }
	}
	else {
	    if (hostname) {
		usage(argv[0]);
		exit(1);
	    }
	    hostname = argv[i];
	}
    }

    if (! hostname) {
#ifdef HOSTFILE
	assert(strlen(argv[0]) < 100);

	/*
	 * Find path of host file from argv[0] and then add name of host
	 * file on the end.
	 */

	strcpy(hostfilename, argv[0]);
	if (strrchr(hostfilename, '/'))
	    strcpy(strrchr(hostfilename, '/') + 1, HOSTFILE);
	else strcpy(hostfilename, HOSTFILE);

	if (hostfile = fopen(hostfilename, "r")) {
	    fscanf(hostfile, "%100s", hostbuf);
	    fclose(hostfile);
	    hostname = hostbuf;
	}
	else {
	    fprintf(stderr,
		 "Host file not found: address of server must be specified\n");
	    usage(argv[0]);
	    exit(0);
	}
#else
	    usage(argv[0]);
	    exit(0);
	}
#endif
    }

    setup_hash();
    connect_to_server(hostname);

    /* Set up alarm signal to occur every second */

    itime.it_interval.tv_sec = 1;
    itime.it_interval.tv_usec = 0;
    itime.it_value.tv_sec = 1;
    itime.it_value.tv_usec = 0;
    signal(SIGALRM, alrm);
    if (setitimer(ITIMER_REAL, &itime, &itime2)) {
	fprintf(stderr, "Error setting interval timer.\n");
	exit(1);
    }

    /* Start off in textmode -- pass text between client and server */

    do {
	do {
	    get_input();
	} while (! (mode & GAMEMODE));

	if (mode & GAMEMODE) {
	    numstart = MAX_PLAYERS;
	    frames = 0;
	    framespersec = 0;
	    ok_to_send = TRUE;
	    no_grenades = MAX_GRENADES;
	    grenade = -1;
	    warp = FALSE;
	    temp = 0;
	    lastenergy = energy = MAX_ENERGY;
	    unit_matrix(&viewer.rotmatrix);
	    unit_matrix(&viewer.inversematrix);

	    viewer.zvel = viewer.yvel = 0L;
	    viewer.xrot = viewer.yrot = 0.0;

	    /* Set all player object pointers to NULL to indicate that
	       no players are currently playing */

	    for (i = 0; i < MAX_LOGINS * 2; i++) playerobj[i] = NULL;
	    world = NULL;  /* No objects have been set up yet */

	    viewer.position.x = viewer.position.z = 0L;
	    viewer.position.y = -CLXHEIGHT;

	    xsize = DEFAULTX;
	    ysize = DEFAULTY;
	    centrex = xsize / 2;
	    centrey = ysize / 2;
	    scale = (float)xsize / tan(VIEWING_ANGLE / 2.0);
	    displ = setup_display(dpyname, xsize, ysize, name1, name2,
		PLANES);
	    /*set_size(DEFAULTX, DEFAULTY);*/
	    cross = XCreateFontCursor(displ->dpy, XC_crosshair);
	    XDefineCursor(displ->dpy, displ->w, cross);
	    set_xcolor(displ, &red, 1.0, 0.0, 0.0);
	    set_xcolor(displ, &black, 0.0, 0.0, 0.0);
	    set_xcolor(displ, &blue, 0.0, 0.0, 1.0);
	    set_xcolor(displ, &skycol, 0.47, 0.69, 1.0); /* Background colour */
	    /* Set up background colour used when player is being shot */
	    set_xcolor(displ, &skycol2, 1.0, 0.33, 0.17);

	    XSetFunction(displ->dpy, displ->gc, GXclear);
	    XCopyArea(displ->dpy, (Drawable)displ->w, (Drawable)displ->pix,
		      displ->gc, 0, 0, xsize, ysize, 0, 0);
	    XCopyArea(displ->dpy, (Drawable)displ->pix, (Drawable)displ->w,
		      displ->gc, 0, 0, xsize, ysize, 0, 0);
	    XSetFunction(displ->dpy, displ->gc, GXcopy);

	    XAutoRepeatOff(displ->dpy);
	    trigger = FALSE;

	    /* Send initial data to server */

	    send_data_to_server();

	    while (mode & GAMEMODE) {
		int entries;

		++frames;
		if (world) {
		    entries = add_sort(world, table, TABLESIZE);
		    for (i = 0; i < MAX_LOGINS * 2; i++) {
			if (playerobj[i])
			    entries += add_sort(playerobj[i], table + entries,
						TABLESIZE - entries);
		    }

		    /* Draw stuff into pixmap */

		    draw_all(displ, table, entries, centrex, centrey, scale,
			&LOS_obj, &collidedata, &portaldata, &teleport);
		    draw_energy_bar();
		    if (no_grenades) draw_grenades();
		    if (LOS_obj) {
			aimobj = LOS_obj;
			while (aimobj->parentobj != world && aimobj->parentobj)
			    aimobj = aimobj->parentobj;
			dist = distance(&aimobj->centre, &viewer.position);
		    }
		    if (LOS_obj && LOS_obj->shoot) {
			draw_sight();
			if (dist < MIN_DIST) draw_sight2();
		    }
		    if (trigger && ok_to_send) temp += 4;
		    if (temp > MAX_TEMP) temp = MAX_TEMP;
		    if (LOS_obj && LOS_obj->shoot && trigger && ok_to_send
		      && temp < MAX_TEMP && dist >= MIN_DIST)
			send_shoot(LOS_obj->shoot);
		    draw_temp_bar();
		    if (trigger && temp < MAX_TEMP) draw_laser();
		    if (temp > 0 && ok_to_send && (! trigger)) temp -= 1;
		    if (collidedata) send_collision(collidedata);
		    if (portaldata) {
			send_collision_portal(portaldata);
			portal_collision = TRUE;
		    }
		    /* Copy pixmap into the actual window */
		    XCopyArea(displ->dpy, (Drawable)displ->pix,
			      (Drawable)displ->w, displ->gc, 0, 0,
			      xsize, ysize, 0, 0);
		 /*   XSetFunction(displ->dpy, displ->gc, GXclear);
		    XCopyArea(displ->dpy, (Drawable)displ->w,
			      (Drawable)displ->pix, displ->gc, 0, 0,
			      xsize, ysize, 0, 0);
		    XSetFunction(displ->dpy, displ->gc, GXcopy);*/
		    /* Clear pixmap */
		    if (energy == lastenergy)  /* We haven't been shot */
			XSetForeground(displ->dpy, displ->gc, skycol.pixel);
		    else if (energy < lastenergy)
			XSetForeground(displ->dpy, displ->gc, skycol2.pixel);
		    else XSetForeground(displ->dpy, displ->gc, blue.pixel);
		    XFillRectangle(displ->dpy, (Drawable)displ->pix,
				   displ->gc, 0, 0, xsize, ysize);
		    /* Make sure this frame is actually drawn on the screen
		       before starting another frame */
		    XSync(displ->dpy, False);
		    /* update_position() is only called if world is defined,
		       so no movement occurs in the absence of reality */
		}

		if (mode & GAMEMODE) {
		    lastenergy = energy;  /* Energy only changes during
					     get_input() */
		    ok_to_send = FALSE;
		    get_input();
		    do_player_clx();
		}

		/* GAMEMODE might not still be set after get_input() */

		if (mode & GAMEMODE) {
		    if (ok_to_send) {
			/* Don't send data to server until it has sent
			   us a packet */
			send_data_to_server();
		    }
		    if (world) trigger = update_position(&map);
		}
	    }
	    XAutoRepeatOn(displ->dpy);
	    XSync(displ->dpy, False);
	    XCloseDisplay(displ->dpy);
	}
    } while (! (mode & CLOSEDOWN));
    close(sckt);
    printf("Connection closed, exiting program.\n");
    return 0;
}  /* main() */

