/*
 * parser.c in Virtual Combat
 * Copyright (c) 1993 by David Jameson
 *
 * Permission to use, copy, modify and distribute this software
 * for any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission appear in the supporting
 * documentation.
 */


/* Routines for parsing an object tree */

#include "defs.h"
#include "3d.h"

#include <math.h>
#include <limits.h>

#define HASHLEN NAMELEN    /* Number of significant characters (needs to
			      be at least NAMELEN) */
#define HASHSIZE 509  /* Size of hash table */
#define WORLDNAME "_world"  /* Name used to identify the WORLDDEF entry
			       in the hash table */

/* Types of hash entry */
#define EMPTY 0        /* No entry exists */
#define COLOURDEF 1    /* Colour definition */
#define POINTDEF 2     /* Points definition */
#define SURFACEDEF 3   /* Surfaces definition */
#define OBJECTDEF 4    /* Object definition */
#define PLAYERDEF 5    /* Player definition */
#define WORLDDEF 6     /* World definition */
#define KEYWORDDEF 7   /* Keyword parsing function */

#define MAXENTRIES 1000  /* Max number of entries in hash table */
#define RECURSION_DEPTH 10  /* Max recursion depth (for build() ) */

#define MAXLINE 10000  /* Maximum length of a script statements */

static char *wspace = " \n\t\v\r\f";  /* White space characters */
static char *wspace2 = " \n\t\v\r\f;";  /* White space characters + semicolon */

/* Module variables */

static char *scriptptr;  /* Pointer into object script */
static object **world;	  /* Root object of world */
static object **playerobj;  /* Array of pointers to player objects */
static point *strt;
static int *strtlen;
static disp *displ;
static int line;  /* Current line number, incremented each time a
		     '\n' is found */



typedef struct _hashentry {
    char name[HASHLEN + 1];   /* Name which was hashed */
    void *dataptr;	      /* Pointer to a function for parsing
				 a keyword or pointer to a data structure */
    struct _hashentry *next;  /* Pointer to another entry which hashes
				 to the same value */
    char hashtype;	      /* Type of hashentry */
} hashentry;  /* Entry in the hash-table */

typedef struct {
    char *keyword;
    void *function;  /* Function for parsing keyword */
} keywordentry;

typedef struct {
    char *name;
    int value;
} flag;

static flag flags[] = {
    { "PLAYER", PLAYER },
    { "ALWAYS_VISIBLE", ALWAYS_VISIBLE },
    { "CHECKX", CHECKX },
    { "CHECKY", CHECKY },
    { "CHECKZ", CHECKZ },
    { "CHECKOBJ", CHECKOBJ },
    { "DRAW_FIRST", DRAW_FIRST },
    { "DRAW_LAST", DRAW_LAST },
    { "END", 0 }
};

typedef struct {
    int num;
    point points[1];
} pointdata;  /* Structure used for storing point data.  The number of points
		 is only necessary for the parser (so that object structures
		 can be properly set up). */

typedef struct {
    int num;
    unsigned short surfaces[1];
} surfdata;  /* Structure used for storing surface data. */

static hashentry hashtable[HASHSIZE];

/* prototypes */

static char *copybuf(char *to, char *from, char term, int n);
static int parse_create(void);
static int parse_position(void);
static int parse_translate(void);
static int parse_rotate_x(void);
static int parse_rotate_y(void);
static int parse_rotate_z(void);
static int parse_colour(void);
static int parse_destroy(void);
static int parse_remove(void);
static int destroy_object(object *searchobj, object *template, object *ob);
static int parse_surfaces(void);
static int parse_points(void);
static int parse_object(void);
static int parse_start(void);
static int parse_world(void);
static int parse_player(void);
static object *build(object *parent, object *tmpl, char *collide,
    char *shoot, int depth, int is_a_player);
static object *search(object *searchobj, object *template, object *ob);
static hashentry *newhash(char *name);
static hashentry *lookup(char *name);
static int hash(char *name);
static char *strtok2(char *s, char *ct);
static int strcount(char *s, char c);
static int hashstat(void);


static int hashstat(void)
{
    int i;
    int entries, ptrs;
    hashentry *ptr;
    int chainsize, maxchainsize;

    entries = ptrs = maxchainsize = 0;

    for (i = 0; i < HASHSIZE; i++) if (hashtable[i].hashtype != EMPTY) {
	chainsize = 0;
	entries++;
	for(ptr = hashtable[i].next; ptr != NULL; ptr = ptr->next) {
	    if (ptr) {
		++ptrs;
		chainsize++;
	    }
	}
	if (chainsize > maxchainsize) maxchainsize = chainsize;
    }
    printf("Size of hash table: %d\n", HASHSIZE);
    printf("Table contains %d entries (%d direct, %d in linked lists)\n",
	entries + ptrs, entries, ptrs);
    printf("Maximum chain length: %d\n", maxchainsize);
    return TRUE;
}  /* hashstat() */


/* Count the number of times the character c occurs in the string s */
static int strcount(char *s, char c)
{
    int count;

    for (count = 0; *s != '\0'; s++) if (*s == c) count++;
    return count;
}  /* strcount() */


/* This version of strtok does the same thing as strtok(3) except that
   the next time it is called (with NULL) it will start scanning from
   after the last token and any separating characters (i.e. separators
   on either side will be ignored instead of just at the beginning).
   Initial white space will be ignored */
static char *strtok2(char *s, char *ct)
{
    static char *prev;
    char *start;
    char *ptr;

    if (! s) s = prev;
    ptr = s;
    while (isspace(*s)) s++;  /* Remove initial white space */
    s += strspn(s, ct);  /* Remove initial separators */
    start = s;
    s = strpbrk(s, ct);  /* Find end of token */
    if (s) {
	prev = s + strspn(s, ct);  /* Remove trailing separators */
	for (; ptr < prev; ptr++) if (*ptr == '\n') line++;
	*s = '\0';
	if (*start) return start;
	else return NULL;
    }
    prev = start;
    for (; ptr < prev; ptr++) if (*ptr == '\n') line++;
    return NULL;
}  /* strtok2() */


/* Find position in hashtable */
static int hash(char *name)
{
    int hashval;

    /* On each iteration swap first and last 8 bits of hashval and then
       exclusive-or with new letter */
    for (hashval = 0; *name != '\0'; name++)
	hashval = (131 * hashval  + *name) % HASHSIZE;

    return (hashval);
}  /* hash() */


/* Find an entry in the hashtable */
static hashentry *lookup(char *name)
{
    hashentry *ptr;

    for (ptr = &hashtable[hash(name)]; ptr != NULL; ptr = ptr->next)
	if (! strcmp(ptr->name, name)) return ptr;
    return NULL;  /* Entry not found */
}  /* lookup() */


/* Find position to put a new entry into the hash table and set up
   pointer to next entry in linked list. */
static hashentry *newhash(char *name)
{
    int hashval;
    hashentry *newentry;
    static entries = 0;

    if (entries >= MAXENTRIES) {
	/* This shouldn't happen */
	fprintf(stderr, "Too many entries in hash table.\n");
	exit(1);
    }
    if (newentry = lookup(name)) {
	if (newentry->hashtype == KEYWORDDEF) {
	    fprintf(stderr, "Error: %s already exists in hash table.\n", name);
	    return NULL;
	}
	else return newentry;
    }
    hashval = hash(name);
    if (hashtable[hashval].hashtype == EMPTY) {
	hashtable[hashval].next = NULL;
	entries++;  /* Incremented only when a new entry is made */
	return &hashtable[hashval];
    }
    /* Place entry at beginning of linked list */
    newentry = (hashentry *)malloc(sizeof(hashentry));
    if (! newentry) return NULL;
    newentry->next = hashtable[hashval].next;
    hashtable[hashval].next = newentry;
    entries++;
    return newentry;
}  /* newhash() */


/* Search through a structure of template objects for a certain object
   and return a pointer to the corresponding real object */
static object *search(object *searchobj, object *template, object *ob)
{
    int i;
    object *tempobj;

    if (! template) return NULL;
    if (! ob) {
	/* This will only happen if "ob" is NULL and "template" is
	   non-NULL.  It should never happen. */
	fprintf(stderr, "search(): object inconsistency.\n");
	return NULL;
    }
    if (searchobj == template) return ob;  /* Found it! */
    for (i = 0; i < template->children; i++)
	if (tempobj = search(searchobj, child_object(template, i),
	    child_object(ob, i))) return tempobj;
    return NULL;  /* Not found */
}  /* search() */


/* Return a pointer to an actual object (not a template) given the name
   of an object */
object *find_object(char *ob)
{
    object *template, *searchobj;
    hashentry *hashptr, *objecthash;

    if (! (objecthash = lookup(ob))) {
	fprintf(stderr, "Line %d, '%s': undefined\n", line, ob);
	return NULL;
    }
    if (objecthash->hashtype == WORLDDEF) return *world;
    if (objecthash->hashtype == PLAYERDEF)
	return *(object **)objecthash->dataptr;
    if (objecthash->hashtype != OBJECTDEF) {
	fprintf(stderr, "Line %d, '%s': not an object\n", line, ob);
	return NULL;
    }

    if (! (hashptr = lookup(WORLDNAME))) {
	fprintf(stderr, "Line %d, can't find object: world not defined\n",
		line);
	return NULL;
    }
    template = (object *)hashptr->dataptr;
    if (! (objecthash = lookup(ob))) {
	fprintf(stderr, "Line %d, '%s': undefined\n", line, ob);
	return NULL;
    }
    searchobj = (object *)objecthash->dataptr;  /* Template being searched
						   for */
    return search(searchobj, template, *world);
}  /* find_object() */


/* Recursively build up a hierarchical list of object structures from
   'template' objects.  These are the actual objects used by the
   3d routines.  A completely new object structure is defined for each
   new definition of an object so that each object can only have one
   parent object (unlike the template objects, which are only defined
   once). */
static object *build(
    object *parent,  /* Pointer to parent object (not template) */
    object *tmpl,    /* pointer to a template object */
    char *collide,   /* Pointer to name of object sent to server when a
			collision occurs */
    char *shoot,     /* Name of object sent when the object gets shot */
    int depth,	     /* Maximum recursion depth */
    int is_a_player) /* Is this object a player? */
{
    object *obj;
    object **childptr;
    int i;

    if (! tmpl) return NULL;
    depth--;
    if (depth < 0) return NULL;

    if (! (obj = (object *)malloc(sizeof(object)))) {
	fprintf(stderr, "Can't allocate memory.\n");
	exit(1);
    }
    if (! (childptr = (object **)malloc(tmpl->children * sizeof(object *)))) {
	fprintf(stderr, "Can't allocate memory.\n");
	exit(1);
    }
    memcpy(obj, tmpl, sizeof(object));
    obj->childobj = childptr;
    if (obj->collide) collide = obj->collide;
    else obj->collide = collide;
    if (obj->shoot) shoot = obj->shoot;
    else obj->shoot = shoot;
    obj->parentobj = parent;

    /*
     * If this object is a player then set the player flag on this object
     * and its child objects.
     */

    if (is_a_player) obj->flags |= PLAYER;

    for (i = 0; i < tmpl->children; i++)
	childptr[i] = build(obj, child_object(tmpl, i), collide, shoot,
			    depth, is_a_player);
    obj->farobj = build(parent, tmpl->farobj, collide, shoot, depth,
		  is_a_player);
    return obj;
}  /* build() */


/* Recursively destroy an object (free the memory allocated for it) */
void destroy(object *obj)
{
    int i;

    if (obj) {
	for (i = 0; i < obj->children; i++) destroy(child_object(obj, i));
	destroy(obj->farobj);
	free((char *)obj->childobj);
	free((char *)obj);
    }
}  /* destroy() */


/* Destroy an object which is a descendent of "object".
   Returns TRUE if the object is found and destroyed. */
static int destroy_object(object *searchobj, object *template, object *ob)
{
    int i;

    if (! template) return FALSE;
    if (! ob) {
	/* This will only happen if "ob" is NULL and "template" is
	   non-NULL.  It should never happen. */
	fprintf(stderr, "destroy_object(): object inconsistency.\n");
	return FALSE;
    }
    if (ob->children != template->children) {
	fprintf(stderr, "destroy_object(): object inconsistency.\n");
	return FALSE;
    }
    for (i = 0; i < template->children; i++)
	if (child_object(template, i) == searchobj) {
	    destroy(child_object(ob, i));
	    delete_child(template, i);
	    delete_child(ob, i);
	    return TRUE;
	}
    /* Two separate loops are used so that objects in the top level of
       the hierarchy will be found more quickly */
    for (i = 0; i < template->children; i++)
	if (destroy_object(searchobj, child_object(template, i),
			   child_object(ob, i))) return TRUE;
    return FALSE;  /* Not found */
}  /* destroy_object() */



/*
 * All parsing routines return TRUE on success and FALSE on failure.
 */


static int parse_player(void)
{
    object *obj;
    hashentry *objecthash;
    int id;
    char *token;
    hashentry *hashptr;
    char *collide, *shoot;

    collide = shoot = NULL;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    hashptr = newhash(token);
    if (hashptr) {
	strncpy(hashptr->name, token, HASHLEN);
	hashptr->name[HASHLEN] = 0;
	hashptr->hashtype = PLAYERDEF;
    }
    else {
	fprintf(stderr, "Can't allocate space for hash table.\n");
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    id = atoi(token);  /* ID number of player */
    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (objecthash = lookup(token))) {
	fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
	return FALSE;
    }
    if (objecthash->hashtype != OBJECTDEF) {
	fprintf(stderr, "Line %d, '%s': not a object\n", line, token);
	return FALSE;
    }
    while (token = strtok2(NULL, wspace2)) {
	if (strcmp(token, "shoot") == 0) shoot = hashptr->name;
	else if (strcmp(token, "collide") == 0) collide = hashptr->name;
	else fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
    }
    obj = build(NULL, (object *)objecthash->dataptr, collide, shoot,
		RECURSION_DEPTH, TRUE);
    playerobj[id] = obj;
    hashptr->dataptr = (void *)&playerobj[id];
    return TRUE;
}  /* parse_player() */


static int parse_world(void)
{
    object *obj;
    hashentry *objecthash;
    char *token;
    hashentry *hashptr;

    hashptr = newhash(WORLDNAME);
    if (hashptr) {
	strncpy(hashptr->name, WORLDNAME, HASHLEN);
	hashptr->name[HASHLEN] = 0;
	hashptr->hashtype = WORLDDEF;
    }
    else {
	fprintf(stderr, "Can't allocate space for hash table.\n");
	return FALSE;
    }
    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (objecthash = lookup(token))) {
	fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
	return FALSE;
    }
    hashptr->dataptr = (void *)objecthash->dataptr;  /* Pointer to world template object */
    if (objecthash->hashtype != OBJECTDEF) {
	fprintf(stderr, "Line %d, '%s': not a object\n", line, token);
	return FALSE;
    }
    obj = build(NULL, (object *)objecthash->dataptr, NULL, NULL,
		RECURSION_DEPTH, FALSE);
    *world = obj;
    return TRUE;
}  /* parse_world() */


/* This routine will read a definition for an object and then build a
   template object structure.  These templates will be read by the
   build() routine, which creates the actual object hierarchy which will
   be used by the 3d routines. */
static int parse_object(void)
{
    object *obj;
    char *token;
    hashentry *hashptr;
    hashentry *pointhash, *surfhash;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (hashptr = lookup(token)) {
	if (hashptr->hashtype != OBJECTDEF) {
	    fprintf(stderr, "Line %d, '%s': symbol already defined\n",
		    line, token);
	    return FALSE;  /* error */
	}
	obj = (object *)hashptr->dataptr;
	if (obj->portal) free(obj->portal);
    }
    else {
	hashptr = newhash(token);
	if (! (obj = (object *)malloc(sizeof(object)))) {
	    fprintf(stderr, "Can't allocate memory.\n");
	    exit(1);
	}
    }
    if (hashptr) {
	strncpy(hashptr->name, token, HASHLEN);
	hashptr->name[HASHLEN] = 0;
	hashptr->hashtype = OBJECTDEF;
    }
    else {
	fprintf(stderr, "Can't allocate space for hash table.\n");
	return FALSE;
    }
    hashptr->dataptr = (void *)obj;

    /* Parse name of point list */

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (pointhash = lookup(token))) {
	fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
	return FALSE;
    }
    if (pointhash->hashtype != POINTDEF) {
	fprintf(stderr, "Line %d, '%s': not a point list\n", line, token);
	return FALSE;
    }
    obj->no_points = ((pointdata *)pointhash->dataptr)->num;
    obj->points = ((pointdata *)pointhash->dataptr)->points;

    /* Parse name of surface list */

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (surfhash = lookup(token))) {
	fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
	return FALSE;
    }
    if (surfhash->hashtype != SURFACEDEF) {
	fprintf(stderr, "Line %d, '%s': not a surface list\n", line, token);
	return FALSE;
    }
    obj->no_surfaces = ((surfdata *)surfhash->dataptr)->num;
    obj->surface = ((surfdata *)surfhash->dataptr)->surfaces;

    obj->flags = 0;
    obj->centre.x = obj->centre.y = obj->centre.z = 0L;
    unit_matrix(&obj->rotmatrix);
    obj->parentobj = obj->farobj = NULL;
    obj->childobj = NULL;
    obj->children = 0;
    obj->shoot = obj->collide = obj->portal = NULL;
    obj->col = NULL;
    obj->light = 1.0;
    obj->maxdist = LONG_MAX;
    while (token = strtok2(NULL, wspace2)) {
	if (strcmp(token, "colour") == 0 || strcmp(token, "color") == 0) {
	    hashentry *colourhash;
	    float light;

	    if (! (token = strtok2(NULL, "\""))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    if (! (colourhash = lookup(token))) {
		fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
		return FALSE;
	    }
	    if (colourhash->hashtype != COLOURDEF) {
		fprintf(stderr, "Line %d, '%s': not a colour definition\n",
			line, token);
		return FALSE;
	    }
	    obj->col = (color *)colourhash->dataptr;
	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    light = (float)atof(token);
	    if (light < 0.0 || light > 1.0) {
		fprintf(stderr, "Line %d, '%s': intensity out of range\n",
			line, token);
		return FALSE;
	    }
	    obj->light = light;
	}
	else if (strcmp(token, "shoot") == 0)  obj->shoot = hashptr->name;
	else if (strcmp(token, "collide") == 0)  obj->collide = hashptr->name;
	else if (strcmp(token, "portal") == 0) {
	    if (! (token = strtok2(NULL, "\""))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }

	    /*
	     * Allocate space for string, which must be freed when object
	     * template is destroyed.
	     */

	    obj->portal = malloc(strlen(token) + 1);
	    strcpy(obj->portal, token);

	    if (! (token = strtok2(NULL, wspace))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    obj->teleport.x = atoi(token);
	    if (! (token = strtok2(NULL, wspace))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    obj->teleport.y = atoi(token);
	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    obj->teleport.z = atoi(token);
	}
	else if (strcmp(token, "child") == 0) {
	    hashentry *childhash;

	    if (! (token = strtok2(NULL, "\""))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    if (! (childhash = lookup(token))) {
		fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
		return FALSE;
	    }
	    if (childhash->hashtype != OBJECTDEF) {
		fprintf(stderr, "Line %d, '%s': not an object definition\n",
			line, token);
		return FALSE;
	    }
	    add_child((object *)childhash->dataptr, obj);
	}
	else if (strcmp(token, "inherit") == 0) {
	    hashentry *copyhash;
	    object *copyobj;
	    int i;

	    if (! (token = strtok2(NULL, "\""))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    if (! (copyhash = lookup(token))) {
		fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
		return FALSE;
	    }
	    if (copyhash->hashtype != OBJECTDEF) {
		fprintf(stderr, "Line %d, '%s': not an object definition\n",
			line, token);
		return FALSE;
	    }
	    copyobj = (object *)copyhash->dataptr;
	    /* Copy properties from inherited object */
	    obj->flags |= copyobj->flags;
	    obj->centre = copyobj->centre;
	    obj->rotmatrix = copyobj->rotmatrix;
	    if (copyobj->col) {
		obj->col = copyobj->col;
		obj->light = copyobj->light;
	    }
	    obj->maxdist = copyobj->maxdist;
	    obj->farobj = copyobj->farobj;
	    /* Copy children from inherited object */
	    for (i = 0; i < copyobj->children; i++)
		add_child(child_object(copyobj, i), obj);
	    if (copyobj->collide) obj->collide = copyobj->collide;
	    if (copyobj->portal) {
		obj->portal = copyobj->portal;
		obj->teleport = copyobj->teleport;
	    }
	    if (copyobj->shoot) obj->shoot = copyobj->shoot;
	}
	else if (strcmp(token, "far_object") == 0) {
	    hashentry *farhash;

	    if (! (token = strtok2(NULL, "\""))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    if (! (farhash = lookup(token))) {
		fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
		return FALSE;
	    }
	    if (farhash->hashtype != OBJECTDEF) {
		fprintf(stderr, "Line %d, '%s': not an object definition\n",
			line, token);
		return FALSE;
	    }
	    obj->farobj = (object *)farhash->dataptr;
	}
	else if (strcmp(token, "maxdist") == 0) {
	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    obj->maxdist = atol(token);
	}
	else if (strcmp(token, "scale") == 0) {
	    float x, y, z;
	    matrix mat;

	    if (! (token = strtok2(NULL, wspace))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    x = (float)atof(token);
	    if (! (token = strtok2(NULL, wspace))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    y = (float)atof(token);
	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    z = (float)atof(token);
	    unit_matrix(&mat);
	    mat.a1 = x;
	    mat.b2 = y;
	    mat.c3 = z;
	    matmult(&mat, &obj->rotmatrix, &obj->rotmatrix);
	}
	else if (strcmp(token, "centre") == 0) {
	    if (! (token = strtok2(NULL, wspace))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    obj->centre.x = atoi(token);
	    if (! (token = strtok2(NULL, wspace))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    obj->centre.y = atoi(token);
	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    obj->centre.z = atoi(token);
	}
	else if (strcmp(token, "rotate_x") == 0) {
	    float angle;

	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    angle = (float)atof(token) * M_PI / 180.0;
	    rotate_x(&obj->rotmatrix, angle, WORLDMODE);
	}
	else if (strcmp(token, "rotate_y") == 0) {
	    float angle;

	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    angle = (float)atof(token) * M_PI / 180.0;
	    rotate_y(&obj->rotmatrix, angle, WORLDMODE);
	}
	else if (strcmp(token, "rotate_z") == 0) {
	    float angle;

	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    angle = (float)atof(token) * M_PI / 180.0;
	    rotate_z(&obj->rotmatrix, angle, WORLDMODE);
	}
	else {
	    flag *ptr;

	    for (ptr = flags; strcmp(ptr->name, "END") !=0; ptr++)
		if (strcmp(token, ptr->name) == 0) {
		    obj->flags |= ptr->value;
		    break;
		}
	    if (strcmp(ptr->name, "END") == 0) {
		fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
		return FALSE;  /* No match */
	    }
	}
    }
    return TRUE;
}  /* parse_object() */


static int parse_points(void)
{
    int psize;	/* Current size of data structure */
    pointdata *p;
    char *token;
    int num;  /* Position in point array */
    hashentry *hashptr;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    psize = sizeof(int) + sizeof(point);
    if (hashptr = lookup(token)) {
	if (hashptr->hashtype != POINTDEF) {
	    fprintf(stderr, "Line %d, '%s': symbol already defined\n",
		    line, token);
	    return FALSE;  /* error */
	}
	if (! (p = (pointdata *)realloc((char *)hashptr->dataptr, psize))) {
	     perror("realloc()");
	     exit(1);
	}
    }
    else {
	hashptr = newhash(token);
	if (! (p = (pointdata *)malloc(psize))) {
	    perror("malloc()");
	    exit(1);
	}
    }
    if (hashptr) {
	strncpy(hashptr->name, token, HASHLEN);
	hashptr->name[HASHLEN] = 0;
	hashptr->hashtype = POINTDEF;
    }
    else {
	fprintf(stderr, "Can't allocate space for hash table.\n");
	return FALSE;
    }
    num = 0;
    do {
	if (token = strtok2(NULL, wspace)) {
	    p->points[num].x = atoi(token);
	    if (! (token = strtok2(NULL, wspace))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    p->points[num].y = atoi(token);
	    if (! (token = strtok2(NULL, wspace2))) {
		fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
		return FALSE;
	    }
	    p->points[num].z = atoi(token);
	    num++;
	    psize = sizeof(int) + sizeof(point) * (num + 1);
	    if (! (p = (pointdata *)realloc((char *)p, psize))) {
		fprintf(stderr, "Can't allocate memory.\n");
		exit(1);
	    }
	}
    } while (token);
    if (num == 0) {
	fprintf(stderr, "Line %d: point list must have at least one point",
		line);
	return FALSE;
    }
    p->num = num;
    hashptr->dataptr = (void *)p;
    return TRUE;
}  /* parse_points() */


static int parse_surfaces(void)
{
    int psize;	/* Current size of data structure */
    surfdata *p;
    char *token;
    int index;	/* Position in surface array */
    int num;  /* Number of surfaces */
    int vertices;  /* Number of vertices for surface */
    hashentry *hashptr;
    int i;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    psize = sizeof(int);
    if (hashptr = lookup(token)) {
	if (hashptr->hashtype != SURFACEDEF) {
	    fprintf(stderr, "Line %d, '%s': symbol already defined\n",
		    line, token);
	    return FALSE;  /* error */
	}
	if (! (p = (surfdata *)realloc((char *)hashptr->dataptr, psize))) {
	     perror("realloc()");
	     exit(1);
	}
    }
    else {
	hashptr = newhash(token);
	if (! (p = (surfdata *)malloc(psize))) {
	    perror("malloc()");
	    exit(1);
	}
    }
    if (hashptr) {
	strncpy(hashptr->name, token, HASHLEN);
	hashptr->name[HASHLEN] = 0;
	hashptr->hashtype = SURFACEDEF;
    }
    else {
	fprintf(stderr, "Can't allocate space for hash table.\n");
	return FALSE;
    }

    num = index = 0;
    do {
	if (token = strtok2(NULL, wspace)) {
	    /* Number of vertices for surface */
	    vertices = (unsigned short)atoi(token);
	    psize += sizeof(unsigned short) * (vertices + 1);
	    num++;
	    if (! (p = (surfdata *)realloc((char *)p, psize))) {
		fprintf(stderr, "Can't allocate memory.\n");
		exit(1);
	    }
	    p->surfaces[index] = vertices;
	    index++;
	    for (i = vertices; i > 0; i--) {
		if (! (token = strtok2(NULL, wspace2))) {
		    fprintf(stderr, "Line %d, '%s': syntax error\n",
			    line, token);
		    return FALSE;
		}
		p->surfaces[index] = (unsigned short)atoi(token);
		index++;
	    }
	}
    } while (token);
    if (num == 0) {
	fprintf(stderr,
		"Line %d: surface list must have at least one surface",
		line);
	return FALSE;
    }
    p->num = num;
    hashptr->dataptr = (void *)p;
    return TRUE;
}  /* parse_surfaces() */


/* Destroy a hierarchical object list (only the objects set up with
   "world ..." or "player ..."; the template objects are not affected) */
static int parse_destroy(void)
{
    char *token, *token2;
    hashentry *hashptr, *worldhash;
    object *template, *searchobj;
    int i;

    if (! (token = strtok2(NULL, "\""))) {
	token2 = strtok2(NULL, wspace2);

	if (token2 && strcmp(token2, "world") == 0) {
	    destroy(*world);
	    *world = NULL;
	    return TRUE;
	}
	else if (token2 && strcmp(token2, "all") == 0) {
	    destroy(*world);
	    *world = NULL;
	    for (i = 0; i < MAX_LOGINS * 2; i++) {
		if (playerobj[i]) {
		    destroy(playerobj[i]);
		    playerobj[i] = NULL;
		}
	    }
	    return TRUE;
	}
	else {
	    fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	    return FALSE;  /* error */
	}
    }
    if (! (hashptr = lookup(token))) {
	fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
	return FALSE;
    }
    switch (hashptr->hashtype) {
	case WORLDDEF :
	    /* Same as parse("destroy world;") */
	    destroy(*world);
	    *world = NULL;
	    break;
	case PLAYERDEF :
	    destroy(*(object **)hashptr->dataptr);
	    *(object **)hashptr->dataptr = NULL;
	    break;
	case OBJECTDEF :
	    if (! (worldhash = lookup(WORLDNAME))) {
		fprintf(stderr,
			"Line %d, can't destroy object: world not defined\n",
			line);
		return FALSE;
	    }
	    template = (object *)worldhash->dataptr;
	    searchobj = (object *)hashptr->dataptr;  /* Template being
							searched for */
	    if (! destroy_object(searchobj, template, *world)) {
		fprintf(stderr,
			"Line %d, '%s': no instances of object found\n",
			line, token);
		return FALSE;
	    }
	    break;
    }
    return TRUE;
}  /* parse_destroy() */


/*
 * Same as parse_destroy except doesn't return an error if no instances
 * of the object are found or if the object isn't defined.
 */
static int parse_remove(void)
{
    char *token, *token2;
    hashentry *hashptr, *worldhash;
    object *template, *searchobj;
    int i;

    if (! (token = strtok2(NULL, "\""))) {
	token2 = strtok2(NULL, wspace2);

	if (token2 && strcmp(token2, "world") == 0) {
	    destroy(*world);
	    *world = NULL;
	    return TRUE;
	}
	else if (token2 && strcmp(token2, "all") == 0) {
	    destroy(*world);
	    *world = NULL;
	    for (i = 0; i < MAX_LOGINS * 2; i++) {
		if (playerobj[i]) {
		    destroy(playerobj[i]);
		    playerobj[i] = NULL;
		}
	    }
	    return TRUE;
	}
	else {
	    fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	    return FALSE;  /* error */
	}
    }
    if (! (hashptr = lookup(token))) {
	return TRUE;
    }
    switch (hashptr->hashtype) {
	case WORLDDEF :
	    /* Same as parse("destroy world;") */
	    destroy(*world);
	    *world = NULL;
	    break;
	case PLAYERDEF :
	    destroy(*(object **)hashptr->dataptr);
	    *(object **)hashptr->dataptr = NULL;
	    break;
	case OBJECTDEF :
	    if (! (worldhash = lookup(WORLDNAME))) {
		fprintf(stderr,
			"Line %d, can't destroy object: world not defined\n",
			line);
		return FALSE;
	    }
	    template = (object *)worldhash->dataptr;
	    searchobj = (object *)hashptr->dataptr;  /* Template being
							searched for */
	    if (! destroy_object(searchobj, template, *world)) {
		return TRUE;
	    }
	    break;
    }
    return TRUE;
}  /* parse_remove() */


/* Set up a colour definition */
static int parse_colour(void)
{
    color *col;
    float r, g, b;
    char *token;
    hashentry *hashptr;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (hashptr = lookup(token)) {
	if (hashptr->hashtype != COLOURDEF) {
	    fprintf(stderr, "Line %d, '%s': symbol already defined\n",
		    line, token);
	    return FALSE;  /* error */
	}
	col = (color *)hashptr->dataptr;
    }
    else {
	hashptr = newhash(token);
	if (! (col = (color *)malloc(sizeof(color)))) {
	    fprintf(stderr, "Can't allocate memory for color structure.\n");
	    exit(1);
	}
    }
    if (hashptr) {
	strncpy(hashptr->name, token, HASHLEN);
	hashptr->name[HASHLEN] = 0;
	hashptr->hashtype = COLOURDEF;
	hashptr->dataptr = (void *)col;
    }
    else {
	fprintf(stderr, "Can't allocate space for hash table.\n");
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    r = (float)atof(token);
    if (r < 0.0 || r > 1.0) {
	fprintf(stderr, "Line %d, '%s': red component out of range\n",
		line, token);
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    g = (float)atof(token);
    if (g < 0.0 || g > 1.0) {
	fprintf(stderr, "Line %d, '%s': green component out of range\n",
		line, token);
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace2))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    b = (float)atof(token);
    if (b < 0.0 || b > 1.0) {
	fprintf(stderr, "Line %d, '%s': blue component out of range\n",
		line, token);
	return FALSE;
    }
    setcol(displ, col, r, g, b);
    return TRUE;  /* No errors */
}  /* parse_colour() */


/* Rotate an object (the actual object is rotated, not the template) */
static int parse_rotate_x(void)
{
    float angle;
    char *token;
    object *obj;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (obj = find_object(token))) {
	fprintf(stderr, "Line %d, '%s': can't find object\n", line, token);
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace2))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    angle = (float)atof(token) * M_PI / 180.0;
    rotate_x(&obj->rotmatrix, angle, WORLDMODE);
    return TRUE;
}  /* parse_rotate_x() */


static int parse_rotate_y(void)
{
    float angle;
    char *token;
    object *obj;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (obj = find_object(token))) {
	fprintf(stderr, "Line %d, '%s': can't find object\n", line, token);
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace2))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    angle = (float)atof(token) * M_PI / 180.0;
    rotate_y(&obj->rotmatrix, angle, WORLDMODE);
    return TRUE;
}  /* parse_rotate_y() */


static int parse_rotate_z(void)
{
    float angle;
    char *token;
    object *obj;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (obj = find_object(token))) {
	fprintf(stderr, "Line %d, '%s': can't find object\n", line, token);
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace2))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    angle = (float)atof(token) * M_PI / 180.0;
    rotate_z(&obj->rotmatrix, angle, WORLDMODE);
    return TRUE;
}  /* parse_rotate_z() */


/* Move an object relative to its current x,y and z coordinates */
static int parse_translate(void)
{
    long dx, dy, dz;
    char *token;
    object *obj;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (obj = find_object(token))) {
	fprintf(stderr, "Line %d, '%s': can't find object\n", line, token);
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    dx = (long)atoi(token);
    if (! (token = strtok2(NULL, wspace))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    dy = (long)atoi(token);
    if (! (token = strtok2(NULL, wspace2))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    dz = (long)atoi(token);

    obj->centre.x += dx;
    obj->centre.y += dy;
    obj->centre.z += dz;

    return TRUE;
}  /* parse_translate() */


/* Move an object to a given position */
static int parse_position(void)
{
    long x, y, z;
    char *token;
    object *obj;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (obj = find_object(token))) {
	fprintf(stderr, "Line %d, '%s': can't find object\n", line, token);
	return FALSE;
    }
    if (! (token = strtok2(NULL, wspace))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    x = (long)atoi(token);
    if (! (token = strtok2(NULL, wspace))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    y = (long)atoi(token);
    if (! (token = strtok2(NULL, wspace2))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;
    }
    z = (long)atoi(token);

    obj->centre.x = x;
    obj->centre.y = y;
    obj->centre.z = z;

    return TRUE;
}  /* parse_position() */


/* Read in starting positions for players.
 * The positions are read into the start[] array.
 */
static int parse_start(void)
{
    char *token;
    int pos;	/* Position in start[] array */
    object *obj;

    pos = 0;
    while (token = strtok2(NULL, wspace2)) {
	if (pos >= *strtlen) {
	    fprintf(stderr, "Line %d, '%s': no room left in array\n",
		line, token);
	    return FALSE;
	}

	strt[pos].x = (long)atoi(token);

	if (! (token = strtok2(NULL, wspace))) {
	    fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	    return FALSE;
	}
	strt[pos].y = (long)atoi(token);
	if (! (token = strtok2(NULL, wspace2))) {
	    fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	    return FALSE;
	}
	strt[pos].z = (long)atoi(token);

	pos++;
    }

    *strtlen -= pos;

    return TRUE;
}  /* parse_start() */



/* Create an object and place it inside the "world".  This is the same
   as defining the object as a child object inside the definition of
   the top-level world object. */
static int parse_create(void)
{
    long x, y, z;
    char *token;
    hashentry *objecthash, *hashptr;
    object *worldtempl;  /* World template object */
    object *obj;
    int translate;  /* Has a new position been specified for the object? */
    object *childobj;

    if (! (token = strtok2(NULL, "\""))) {
	fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	return FALSE;  /* error */
    }
    if (! (objecthash = lookup(token))) {
	fprintf(stderr, "Line %d, '%s': undefined\n", line, token);
	return FALSE;
    }
    if (objecthash->hashtype != OBJECTDEF) {
	fprintf(stderr, "Line %d, '%s': not a object\n", line, token);
	return FALSE;
    }
    obj = (object *)objecthash->dataptr;
    translate = FALSE;
    /* Check to see if optional x,y and z coordinates have been specified */
    if (token = strtok2(NULL, wspace2)) {
	translate = TRUE;
	x = (long)atoi(token);
	if (! (token = strtok2(NULL, wspace))) {
	    fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	    return FALSE;
	}
	y = (long)atoi(token);
	if (! (token = strtok2(NULL, wspace2))) {
	    fprintf(stderr, "Line %d, '%s': syntax error\n", line, token);
	    return FALSE;
	}
	z = (long)atoi(token);
    }
    if (! (hashptr = lookup(WORLDNAME))) {
	fprintf(stderr, "Line %d, can't create object: world not defined\n",
		line);
	return FALSE;
    }
    worldtempl = (object *)hashptr->dataptr;
    /* Add object to the list of child objects inside the world template
       object */
    add_child(obj, worldtempl);
    /* Do the same for the actual world object */
    childobj = build(*world, obj, NULL, NULL, RECURSION_DEPTH, FALSE);
    if (translate) {
	childobj->centre.x = x;
	childobj->centre.y = y;
	childobj->centre.z = z;
    }
    add_child(childobj, *world);
    return TRUE;
}  /* parse_create() */


/* Copies data into buffer until either terminator or '\0' reached and
   returns a pointer to the last character copied.
   Comments are blanked out.
   Resulting string will be NULL terminated. */
static char *copybuf(char *to, char *from, char term, int n)
{
    while (*from != '\0'  &&  *from != term  &&  n > 1) {
	if (*from == '"') {
	    /*
	     * Terminator character and comment characters don't have
	     * significance inside strings.
	     */
	    *to++ = *from++;
	    n--;
	    while (*from != '"' && *from != '\0' && n > 1) {
		*to++ = *from++;
		n--;
	    }
	    if (*from == '\0' || n <= 1) break;
	    *to++ = *from++;
	    n--;
	    continue;
	}
	if ((*from == '/') && (from[1] == '*')) {
	    from += 2;
	    while ((*from != '*' || from[1] != '/') && *from != '\0' &&
		   n > 1) {
		if (*from == '\n') {  /* Leave in line-feeds */
		    *to++ = '\n';
		    n--;
		}
		from++;
	    }
	    if (*from == '\0') break;
	    from += 2;
	    continue;
	}
	if (*from == '#') {
	    while (*from != '\n' && *from != '\0') from++;
	    continue;
	}
	*to++ = *from++;
	n--;
    }
    if (n <= 1) {
	fprintf(stderr, "Buffer overflow: line too long.\n");
	exit(1);
    }
    *to++ = *from;
    *to = '\0';
    return from;
}  /* copybuf() */


void setup_hash(void)
{
    int i;
    keywordentry *ptr;
    hashentry *hashptr;
    static keywordentry keys[] = {
	{ "colour", (void *)parse_colour },
	{ "color", (void *)parse_colour },
	{ "points", (void *)parse_points },
	{ "surfaces", (void *)parse_surfaces },
	{ "object", (void *)parse_object },
	{ "player", (void *)parse_player },
	{ "world", (void *)parse_world },
	{ "destroy", (void *)parse_destroy },
	{ "remove", (void *)parse_remove },
	{ "start", (void *)parse_start },
	{ "hashstat", (void *)hashstat },
	{ "rotate_x", (void *)parse_rotate_x },
	{ "rotate_y", (void *)parse_rotate_y },
	{ "rotate_z", (void *)parse_rotate_z },
	{ "translate", (void *)parse_translate },
	{ "position", (void *)parse_position },
	{ "create", (void *)parse_create },
	{ "END", NULL }
    };

    /* Clear table */

    for (i = 0; i < HASHSIZE; i++) hashtable[i].hashtype = EMPTY;

    for (ptr = keys; strcmp(ptr->keyword, "END") != 0; ptr++) {
	hashptr = newhash(ptr->keyword);
	if (hashptr) {
	    strncpy(hashptr->name, ptr->keyword, HASHLEN);
	    hashptr->name[HASHLEN] = 0;
	    hashptr->dataptr = ptr->function;
	    hashptr->hashtype = KEYWORDDEF;
	}
	else {
	    fprintf(stderr, "Can't allocate space for hash table.\n");
	    return;
	}
    }
}


/* Parse an object script */
int parse(
    disp *d,
    char *script,      /* NULL terminated string */
    point *start,      /* Starting positions */
    int *startlen,     /* Max number of starting positions */
    object **wrld,     /* Root object of world */
    object **players)  /* playerobj set to this */
{
    char linebuf[MAXLINE];
    hashentry *ptr;
    int errors;  /* Number of errors encountered */
    char *token;

    displ = d;
    strt = start;
    strtlen = startlen;
    playerobj = players;
    world = wrld;
    errors = 0;
    line = 1;
    do {
	scriptptr = script;
	script = copybuf(linebuf, scriptptr, ';', MAXLINE);
	/*nl = strcount(linebuf, '\n');*/
	if (*script != '\0') {
	    token = strtok2(linebuf, wspace2);  /* Tokenise keyword */
	    if (! (ptr = lookup(token)))
		fprintf(stderr, "Line %d, '%s': unrecognised keyword\n",
			line, linebuf);
	    else {
		if (ptr->hashtype != KEYWORDDEF)
		    fprintf(stderr, "Line %d, '%s': not a keyword\n",
			    line, linebuf);
		/* Call parsing function for this keyword */
		else {
#if DEBUG >= 2
		    fprintf(stderr, "Parsing keyword \"%s\"\n", token);
#endif
		    if (! (*(int (*)())ptr->dataptr)() ) ++errors;
		    else if (strtok2(NULL, wspace2))
			fprintf(stderr,
    "Line %d: data at end of line ignored\n", line);
		}
	    }
	}
	else if (strtok2(linebuf, wspace2)) fprintf(stderr,
    "Line %d: no semicolon at end of line\n", line);

	/*line += nl;*/
    } while (*script++ != '\0');
    if (errors) printf("Parsed with %d errors.\n", errors);
    return errors;
}  /* parse() */

